from pathlib import Path

import imagesize
import pandas as pd

from ..dataset import Dataset


def from_mot(
    ann_txt: Path | str,
    images_folder: Path | str,
    category_id: int,
    category_str: str,
    split: str | None = None,
) -> Dataset:
    """Load a dataset stored in the MOT format.

    See `specifications <https://motchallenge.net/instructions/>`_

    Note:
        The image filenames must represent the image's id (e.g. 0001.jpg). This id
        is used in the .txt annotation file to associate annotations with this image

    Args:
        ann_txt: path to the .txt file containing the MOT annotations.
        images_folder: path to the folder containing the dataset's images which must
            be at the root of this folder.
        category_id: category_id of the objects that are annotated in your MOT dataset.
            this means that your dataset contains only one class, which is the case of
            the MOT datasets we've tested so far (MOT20Det, CroHD).
        category_str: category_str of the objects annotated in your MOT dataset.
        split: split of the loaded dataset.

    Returns:
        Loaded dataset object
    """
    # convert path arguments to Path objects in case they are strings
    ann_txt = Path(ann_txt)
    images_folder = Path(images_folder)

    # load .txt annotations into a dataframe
    ann_df = ann_txt_to_df(ann_txt)
    # add category_id and category_str columns to annotations df
    ann_df["category_id"], ann_df["category_str"] = category_id, category_str
    # drop useless columns from annotations df
    ann_df.drop(columns=["x_world", "y_world"], inplace=True)

    # create images dataframe
    img_df = pd.DataFrame(
        data=[
            [
                int(img_path.stem),  # image id
                Path(img_path.name),  # image name
                img_path,  # image name
                img_path.suffix,  # image type
                split,
            ]
            for img_path in images_folder.glob("*.jpg")
        ],
        columns=["id", "relative_path", "absolute_path", "type", "split"],
    ).set_index("id")

    # add image width and height columns to images df
    img_df["width"] = img_df["absolute_path"].apply(lambda x: imagesize.get(x)[0])
    img_df["height"] = img_df["absolute_path"].apply(lambda x: imagesize.get(x)[1])
    img_df.drop(columns=["absolute_path"], inplace=True)

    # create the dataset's labelmap
    labelmap = {category_id: category_str}

    return Dataset(
        images_root=images_folder,
        annotations=ann_df,
        images=img_df,
        label_map=labelmap,
    )


def ann_txt_to_df(ann_txt: Path | str) -> pd.DataFrame:
    """Read the .txt MOT annotations into a dataframe.

    Args:
        ann_txt: path to .txt annotation of your original MOT dataset
    Returns:
        pd.DataFrame: a dataframe containing the MOT annotations
    """
    # read the .txt annotations into a dataframe.
    ann_df = pd.read_csv(
        ann_txt,
        names=[
            "image_id",
            "obj_id",
            "box_x_min",
            "box_y_min",
            "box_width",
            "box_height",
            "confidence",
            "x_world",
            "y_world",
        ],
        dtype={
            "image_id": int,
            "obj_id": int,
            "box_x_min": float,
            "box_y_min": float,
            "box_width": float,
            "box_height": float,
            "confidence": float,
        },
    )

    # use the dataframe's index as the id of annotations
    ann_df.index.name = "id"

    # Ignore annotations that have a conf=0 as stated in the MOT's webpage:
    # https://motchallenge.net/instructions/
    ann_df = ann_df[ann_df.confidence != 0]

    # drop confidence col if all rows are equal to 1 (dataset has only groundtruths)
    if (ann_df["confidence"] == 1.0).all():
        ann_df.drop(columns=["confidence"], inplace=True)
    # replace confidence values that equals 1 (used for groundtruths) with NaN
    else:
        ann_df["confidence"][ann_df["confidence"] == 1.0] = float("nan")

    return ann_df



from .schema_util_functions import (
    fill_with_dtypes_and_default_value,
    flatten_schema,
    get_dtypes_and_default_values,
    get_enums,
    get_remapping_dict_from_names,
    get_remapping_dict_from_schema,
    load_json_schema,
    remap_dict,
)

__all__ = [
    "get_dtypes_and_default_values",
    "get_enums",
    "get_remapping_dict_from_names",
    "get_remapping_dict_from_schema",
    "flatten_schema",
    "load_json_schema",
    "fill_with_dtypes_and_default_value",
    "remap_dict",
]



from collections.abc import Iterable
from pathlib import Path
from shutil import rmtree

import pandas as pd
from yaml import safe_dump, safe_load

from lours import __version__


def dict_to_parquet(
    output_dict: dict,
    output_dir: Path,
    version: str = __version__,
    fields: Iterable[str] | None = None,
    fields_to_str: Iterable[str] = ("images_root", "relative_path", "images_root"),
    overwrite: bool = False,
) -> None:
    """Save a dictionary containing dataframes as a yaml file and parquet files.

    The dataset can be nested.

    Args:
        output_dict: dictionary to save, containing yaml serializable objects and
            dataframes. Can be nested.
        output_dir: path to folder where to save the yaml and parquets files
        version: data version info for future compatibility.
            Defaults to current Lours version.
        fields: fields to save. Will ignore other fields in the output dictionary.
            If set to None, will save all fields. Defaults to None.
        fields_to_str: fields to convert to str. Useful for non-serializable objects
            like Path
        overwrite: if set to True, will remove the ``output_dir`` directory if it
            already exists. If set to False, will check that the directory either does
            not exist or is empty. Defaults to False

    Raises:
        OSError: Raised when the output directory is not empty and ``overwrite`` is set
            to False
    """
    if overwrite:
        rmtree(output_dir, ignore_errors=True)
        output_dir.mkdir(parents=True)
    elif output_dir.is_dir():
        if any(output_dir.iterdir()):
            raise OSError("Output directory must be empty")
    else:
        output_dir.mkdir(parents=True)

    def replace_dataframes_in_dict(dict_to_convert, prefix):
        converted_dict = {}
        for name, attribute in dict_to_convert.items():
            if fields is not None and name not in fields:
                continue
            if isinstance(attribute, pd.DataFrame):
                parquet_name = f"{prefix}{name}.pq"
                output_parquet_file = output_dir / parquet_name
                converted_dict[name] = f"DataFrame:{parquet_name}"
                attribute.astype(
                    {f: str for f in fields_to_str if f in attribute.columns}
                ).to_parquet(output_parquet_file)
            elif isinstance(attribute, dict):
                converted_dict[name] = replace_dataframes_in_dict(
                    attribute, f"{prefix}{name}."
                )
            else:
                converted_dict[name] = (
                    attribute if name not in fields_to_str else str(attribute)
                )
        return converted_dict

    metadata = replace_dataframes_in_dict(output_dict, "")
    metadata["version"] = version
    with open(output_dir / "metadata.yaml", "w") as f:
        safe_dump(metadata, f)


def dict_from_parquet(
    input_dir: Path,
    fields_to_path: Iterable[str] = ("relative_path", "images_root"),
) -> dict:
    """Create dictionary from folder created with the function :func:`.dict_to_parquet`

    Args:
        input_dir: folder containing yaml and parquet files
        fields_to_path: Iterable of strings to specify which columns will need to be
            converted to Path objects.

    Returns:
        created dictionary. Will be used to reconstruct objects with dataframes, such
        as Dataset or Evaluator.
    """

    def replace_dataframe_placeholders(input_dict):
        for name, attribute in input_dict.items():
            if isinstance(attribute, str) and attribute.startswith("DataFrame:"):
                parquet_path = input_dir / attribute.split(":")[1]
                loaded_dataframe = pd.read_parquet(parquet_path)
                for f in fields_to_path:
                    if f in loaded_dataframe.columns:
                        loaded_dataframe[f] = loaded_dataframe[f].apply(
                            Path  # pyright: ignore
                        )
                input_dict[name] = loaded_dataframe

            elif isinstance(attribute, dict):
                replace_dataframe_placeholders(attribute)

    with open(input_dir / "metadata.yaml") as f:
        metadata = safe_load(f)

    for f in fields_to_path:
        if f in metadata:
            metadata[f] = Path(metadata[f])

    replace_dataframe_placeholders(metadata)
    return metadata



"""Set of tools to differentiate datasets or evaluators"""

from collections.abc import Iterable
from warnings import warn

from lours.dataset import Dataset
from lours.utils.label_map_merger import IncompatibleLabelMapsError, merge_label_maps


def dataset_diff(
    left_dataset: Dataset,
    right_dataset: Dataset,
    exclude_image_columns: Iterable[str] = (),
    exclude_annotations_columns: Iterable[str] = (),
) -> tuple[Dataset, Dataset, Dataset]:
    """Differentiate two datasets and construct the difference datasets, only containing
    elements that are in one of the two datasets but not the other

    this function outputs the differences with 2 datasets that are constructed with
    images and annotations specific to each dataset and a third dataset with common
    images and annotations.

    As such, you should theoretically be able to reconstruct the left dataset with
    the first difference dataset and the common dataset, and reconstruct the right
    dataset with the second difference dataset and the common dataset.

    Note:
        if one dataset has a column in its dataframes the other dataset doesn't have,
        and that column is not included in ``exclude_image_columns`` or
        ``exclude_annotations_column``, the dataframes and thus the datasets will be
        considered entirely different, and the common dataset will be empty

    Note:
        if ``exclude_image_columns`` or ``exclude_annotations_columns`` is not empty,
        it is not guaranteed to be able to reconstruct left or right dataset with common
        datasets and difference datasets, only the datasets minus the excluded columns.


    Args:
        left_dataset: left dataset to compare
        right_dataset: right dataset to compare
        exclude_image_columns: list of names of columns to ignore in image dataframes
            for the comparison.
        exclude_annotations_columns: list of names of columns to ignore in annotations
            dataframes for the comparison.

    Returns:
        tuple with 3 datasets
         - dataset with images and annotations that are specific to ``left_dataset``
         - dataset with images and annotations that are specific to ``right_dataset``
         - dataset with images and annotations that are common to both input datasets.
    """
    try:
        merge_label_maps(left_dataset.label_map, right_dataset.label_map)
    except IncompatibleLabelMapsError:
        warn("Incompatible label maps, dataset cannot have mutual info", RuntimeWarning)
        return left_dataset, right_dataset, Dataset()

    exclude_annotations_columns = list(exclude_annotations_columns)
    exclude_image_columns = list(exclude_image_columns)
    left_images, left_annotations = left_dataset.images, left_dataset.annotations
    right_images, right_annotations = right_dataset.images, right_dataset.annotations
    left_images = left_images.drop(columns=exclude_image_columns, errors="ignore")
    right_images = right_images.drop(columns=exclude_image_columns, errors="ignore")
    left_annotations = left_annotations.drop(
        columns=exclude_annotations_columns, errors="ignore"
    )
    right_annotations = right_annotations.drop(
        columns=exclude_annotations_columns, errors="ignore"
    )

    if exclude_annotations_columns or exclude_image_columns:
        # Construct new datasets with the excluded columns removed.
        # these one shoulb have the exact same columns
        return dataset_diff(
            left_dataset.loc[:, left_images.columns].loc_annot[
                :, left_annotations.columns
            ],
            right_dataset.loc[:, right_images.columns].loc_annot[
                :, right_annotations.columns
            ],
        )

    # If datasets are equal, just output left dataset with the excluded columns
    if left_images.equals(right_images) and left_annotations.equals(right_annotations):
        return left_dataset.loc[[]], right_dataset.loc[[]], left_dataset

    if set(left_images.columns) != set(right_images.columns):
        warn(
            "Column mismatch between Image DataFrames, consider using"
            " 'exclude_image_columns' argument, or comparing after manually excluding"
            " not shared columns in your datasets images frames",
            RuntimeWarning,
        )
        return left_dataset, right_dataset, Dataset()

    if set(left_annotations.columns) != set(right_annotations.columns):
        warn(
            "Column mismatch between Annotations DataFrames, consider using"
            " 'exclude_annotations_columns' argument, or comparing after manually"
            " excluding not shared columns in your datasets annotations frames",
            RuntimeWarning,
        )
        return left_dataset, right_dataset, Dataset()

    left_image_ids = set(left_images.index)
    right_image_ids = set(right_images.index)

    common_image_ids = left_image_ids & right_image_ids
    common_image_ids_list = list(common_image_ids)
    only_left_image_ids = left_image_ids - right_image_ids
    only_right_image_ids = right_image_ids - left_image_ids

    # compare images that share the same id across both datasets
    common_left_images = left_images.loc[common_image_ids_list]
    common_right_images = right_images.loc[common_image_ids_list]

    changed_values = common_left_images != common_right_images
    # None values should be considered the same here, the same pandas does with the
    # `equals` method
    changed_values[common_left_images.isna() & common_left_images.isna()] = False
    changed_images = changed_values.any(axis=1)
    changed_images_ids = set(common_left_images.loc[changed_images].index)

    left_annotations_ids = set(
        left_dataset.loc[common_image_ids_list].annotations.index
    )
    right_annotations_ids = set(
        right_dataset.loc[common_image_ids_list].annotations.index
    )
    changed_images_annotations_ids = set(
        left_dataset.loc[list(changed_images_ids)].annotations.index
    ) | set(right_dataset.loc[list(changed_images_ids)].annotations.index)

    # annotations that are the same but are linked to images that are different
    # are not considered the same
    common_annotations_ids = (
        left_annotations_ids & right_annotations_ids
    ) - changed_images_annotations_ids
    common_annotations_ids_list = list(common_annotations_ids)
    only_left_annotations_ids = left_annotations_ids - right_annotations_ids
    only_right_annotations_ids = right_annotations_ids - left_annotations_ids

    # Now compare annotations that share the same id and are linked to images that
    # are the same across datasets
    common_left_annot = left_annotations.loc[common_annotations_ids_list]
    common_right_annot = right_annotations.loc[common_annotations_ids_list]
    changed_annotations_values = common_left_annot != common_right_annot
    # Same as for images, None values in both frames should be considered the same
    changed_annotations_values[common_left_annot.isna() & common_right_annot.isna()] = (
        False
    )
    changed_annotations = changed_annotations_values.any(axis=1)
    changed_annotations_ids = set(common_left_annot.loc[changed_annotations].index)

    # resulting left dataset is composed of:
    # - images that are only in left dataset and all their annotations
    # - images that are different from right dataset and all
    #   their annotations
    # - annotation that are only in left dataset and their linked image
    # - annotations that are different from right dataset and their linked image
    only_left_dataset = (
        left_dataset.loc[list(only_left_image_ids | changed_images_ids)]
        + left_dataset.loc_annot[
            list(only_left_annotations_ids | changed_annotations_ids)
        ].remove_empty_images()
    )

    # Same applied for right dataset
    only_right_dataset = (
        right_dataset.loc[list(only_right_image_ids | changed_images_ids)]
        + right_dataset.loc_annot[
            list(only_right_annotations_ids | changed_annotations_ids)
        ].remove_empty_images()
    )

    # Resulting common dataset is composed of images that are in both datasets and
    # that are the same across datasets, and only their annotations if they are
    # in both datasets and are the same across datasets.
    common_dataset = (
        left_dataset.loc[list(common_image_ids - changed_images_ids)]
        .loc_annot[list(common_annotations_ids - changed_annotations_ids)]
        .remove_empty_images()
    )

    return only_left_dataset, only_right_dataset, common_dataset



from collections.abc import Iterable

import pandas as pd
from ipykernel.zmqshell import ZMQInteractiveShell
from IPython.core.getipython import get_ipython
from IPython.display import display
from ipywidgets import (
    Dropdown,
    HBox,
    Label,
    Layout,
    Output,
    Stack,
    ToggleButtons,
    VBox,
    jslink,
)
from pandas.io.formats.style_render import CSSDict

from .column_booleanizer import debooleanize

is_notebook = isinstance(get_ipython(), ZMQInteractiveShell)


def convert_columns_to_nested_multi_index(
    columns: "pd.Index[str]", separator: str = "."
) -> pd.MultiIndex:
    """Convert an index with names containing a particular separator
    to a multiIndex. Names without separator will have blank index values for
    level above 0.

    Note:
        This is only for html displaying purpose, the multiIndex is not meant to be
        meaningful

    Args:
        columns: pandas string index of columns names
        separator: separator character to use to split column names into index levels.
            Defaults to ".".

    Returns:
        Multi Index with has many levels as the maximum number of separator character in
        a column name.
    """
    max_nest_level = 1
    nested_names = []
    for name in columns:
        names = name.split(separator)
        max_nest_level = max(max_nest_level, len(names))
        nested_names.append(names)

    def pad_tuple(input_names: list[str]) -> tuple[str, ...]:
        pad_size = max_nest_level - len(input_names)
        return tuple(input_names + [""] * pad_size)

    return pd.MultiIndex.from_tuples([pad_tuple(names) for names in nested_names])


def display_booleanized_dataframe(
    input_df: pd.DataFrame,
    booleanized_columns: Iterable[str] = (),
    separator: str = ".",
) -> None:
    """Utilitary function to display a dataset dataframe properly.

    More specifically, this will try to make easier to read nested data and booleanized
    data.

    In the case data is booleanized or nested, add some interface to let the user chose
    between raw display (the dataframe as it is present in the python code) or formatted
    display, by debooleanizing or transforming the columns into multi index.

    Args:
        input_df: Input dataframe to display. Its column names are potentially very long
            because of booleanized or nested data structure.
        booleanized_columns: List of prefix to search for booleanized columns. The
            ``input_df`` dataframe will be debooleanized for displaying purpose.
            Defaults to ().
        separator: character used to indicate nested data. Will be used for
            debooleanization and column name splitting.
    """
    from . import DISPLAY_NESTED_COLUMNS, DISPLAY_UNBOOLEANIZED

    booleanized_columns = list(booleanized_columns)
    nothing_to_format = not any("." in name for name in input_df.columns)
    if nothing_to_format:
        # No booleanized data, and no nested data. Nothing to format, simply output the
        # dataframe as raw
        display(input_df)
        return
    # Debooleanization can be expensive. Only use the head and tail of dataframe to
    # make this operation cheaper, since it's only for displaying purpose
    if len(input_df) > 100:
        short_df = pd.concat([input_df.iloc[:50], input_df.iloc[:-50]])
    else:
        short_df = input_df.copy()

    # Additional style for nested multiIndex : add dim vertical lines to keep track
    # of the tree structure
    # Note that the type hint is not really necessary, but pyright complains without
    # it. See
    # https://github.com/microsoft/pyright/blob/main/docs/type-concepts.md#generic-types
    nested_table_styles: list[CSSDict] = [
        {"selector": "th", "props": "border-right: 1px solid #F0F0F0"}
    ]

    short_df_nested = short_df.copy()
    short_df_nested.columns = convert_columns_to_nested_multi_index(
        short_df.columns,
        separator=separator,
    )
    short_df_nested = short_df_nested.style.set_table_styles(nested_table_styles)

    short_df_debooleanized = debooleanize(
        short_df, booleanized_columns, separator=separator
    )
    short_df_debooleanized_nested = short_df_debooleanized.copy()
    short_df_debooleanized_nested.columns = convert_columns_to_nested_multi_index(
        short_df_debooleanized.columns, separator=separator
    )
    short_df_debooleanized_nested = (
        short_df_debooleanized_nested.style.set_table_styles(nested_table_styles)
    )

    dataframe_outputs = {
        name: Output() for name in ["raw", "nested", "debool_raw", "debool_nested"]
    }
    with dataframe_outputs["raw"]:
        display(short_df)
    with dataframe_outputs["nested"]:
        display(short_df_nested)
    with dataframe_outputs["debool_raw"]:
        display(short_df_debooleanized)
    with dataframe_outputs["debool_nested"]:
        display(short_df_debooleanized_nested)

    default_dropdown_value = "nested" if DISPLAY_NESTED_COLUMNS else "raw"
    column_format_select = Dropdown(
        options=["raw", "nested"], value=default_dropdown_value
    )

    if booleanized_columns:
        bool_stack = Stack([dataframe_outputs["raw"], dataframe_outputs["nested"]])
        jslink((column_format_select, "index"), (bool_stack, "selected_index"))
        unbool_stack = Stack(
            [
                dataframe_outputs["debool_raw"],
                dataframe_outputs["debool_nested"],
            ]
        )
        jslink(
            (column_format_select, "index"),
            (unbool_stack, "selected_index"),
        )
        # jslink does not work with link between true/false and int index,
        # So we have to get rid of the checkbox to use a toggle buttons instead.
        # See related issue. Given how old it is, we might not have a solution in the
        # near future
        # https://github.com/jupyter-widgets/ipywidgets/issues/1109
        bool_toggle = ToggleButtons(
            options=["yes ", "no "],
            icons=["check", "times"],
            layout=Layout(width="auto"),
            style={"button_width": "auto"},
            value="no " if DISPLAY_UNBOOLEANIZED else "yes ",
        )
        selector = HBox(
            [
                VBox([Label("Booleanize"), Label("Column format")]),
                VBox([bool_toggle, column_format_select]),
            ]
        )
        stack = Stack([bool_stack, unbool_stack])
        jslink((bool_toggle, "index"), (stack, "selected_index"))
    else:
        # No booleanized columns, no point in displaying the booleanized checkbox
        selector = HBox(
            [
                Label("Column format"),
                column_format_select,
            ]
        )
        stack = Stack([dataframe_outputs["raw"], dataframe_outputs["nested"]])
        jslink((column_format_select, "index"), (stack, "selected_index"))

    display(VBox([selector, stack]))



from collections.abc import Sequence
from pathlib import Path
from typing import Literal, TypeGuard

import numpy as np
import pandas as pd
from faker import Faker

from ..dataset import Dataset
from .bbox_converter import import_bbox


def construct_attribute_column(
    numpy_generator: np.random.Generator,
    n_rows: int,
    labels: Sequence[str],
    probs: Sequence[float] | None = None,
    is_list_column: bool = True,
) -> pd.Categorical | list[list[str]]:
    """Generate a column with lists of elements taken in a finite pool
    the generated sequence of lists will be in the form of a numpy array, which will
    become a column in a DataFrame.

    Args:
        numpy_generator: numpy random Generator object used to generate random integers
        n_rows: number of rows of generated numpy array
        labels: label strings to use for the attributes
        probs: sequence of probabilities to construct each row. If set to None, will use
            the probabilities by default: for attribute lists, each probability will
            be 0.5, and for simple attribute, probabilities will be evenly distributed.
            Defaults to None.
        is_list_column: if set to True, will construct a column with list of attributes,
            that constitute a subset of the set of labels. Otherwise, will simply
            construct a simple attribute column, where each row is a single label taken
            from ``labels`` according to the probability distribution given by
            ``probs``. Defaults to True.

    Returns:
        list of lists that will be incorporated in a dataframe.
    """
    if is_list_column:
        if probs is None:
            booleanized_column = numpy_generator.integers(
                0, 2, (n_rows, len(labels))
            ).astype(bool)
        else:
            booleanized_column = np.stack(
                list(
                    numpy_generator.choice([True, False], size=n_rows, p=[p_, 1 - p_])
                    for p_ in probs
                ),
                axis=-1,
            )
        labels_np = np.array(list(labels))
        return [list(labels_np[indices]) for indices in booleanized_column]
    else:
        return pd.Categorical(
            numpy_generator.choice(labels, size=n_rows, p=probs), categories=labels
        )


random_attribute_column_type = (
    int
    | Sequence[int]
    | Sequence[str]
    | Sequence[Sequence[float]]
    | dict[str, int]
    | dict[str, Sequence[float]]
    | dict[str, Sequence[str]]
    | dict[str, dict[str, float]]
)
"""The random attribute columns type is a way to design a column with random
attributes.

It will create :math:`N` columns, each :math:`i` th column with :math:`M_i` labels,
the labels being distributed according to the probabilities in :math:`(p_i)_j`
(:math:`(p_i)_j` being of length :math:`M_i`, with values :math:`p_{i,j}` between
0 and 1).

In the case the column is an non-list attribute column, each vector :math:`(p_i)_j` must
addup to 1. Otherwise, each probability :math:`p_{i,j}` is the probability that the
:math:`j` th label of :math:`i` th column is in the attribute list for each cell.

Depending on the type, the values :math:`N`, :math:`M`, :math:`(p_i)_j` and the names
will be constructed differently.

If not specified, column header and labels are randomly generated with
``Faker.unique.word()``

If not specified, the probabilities pi will be either uniform probabilities for
non-list attribute columns, or all set to 0.5 for attribute list columns.

The input can be either

- An integer: :math:`N` is the given integer, :math:`M_i` are random integers between
  2 and 10
- A sequence of integers: :math:`N` is the length of the sequence, :math:`M_i` are the
  integers of that sequence.
- A sequence of str: :math:`N` is the length of the sequence, the column headers are the
  sequence elements, and :math:`M_i` are random integers between 2 and 10.
- A sequence of sequences of float: :math:`N` is the length of the sequence,
  :math:`M_i` is the length of each :math:`i` th sequence, and :math:`(p_i)_j` is the
  :math:`i` th sequence of floats.
- A dictionary of integers: :math:`N` is the length of the dictionary. The column
  headers are the dictionary keys, and :math:`M_i` are the integer values.
- A dictionary of float sequences: :math:`N` is the length of the dictionary. The column
  headers are the dictionary keys, :math:`M_i` is the length of the :math:`i` th float
  sequence, and :math:`(p_i)_j` is the :math:`i` th float sequence
- A dictionary of string sequences: :math:`N` is the length of the dictionary. The
  column headers are the dictionary keys, :math:`M_i` is the length of the :math:`i` th
  string sequence, and the :math:`j` th label of the :math:`i` th column is the
  :math:`j` th element of the :math:`i` th sequence.
- A dictionary of float dictionaries. :math:`N` is the length of the root dictionary.
  The column headers are the dictionary keys, :math:`M_i` is the length of the
  :math:`i` th sub-dictionary,  the :math:`j` th label of the :math:`i` th column is the
  :math:`j` th key of the :math:`i` th sub-dictionary and the probability
  :math:`p_{i,j}` is the corresponding sub-dictionary value
"""


def set_attribute_columns_labels(
    input_dataframe: pd.DataFrame,
    columns_specs: random_attribute_column_type,
    numpy_generator: np.random.Generator,
    fake_generator: Faker,
    is_list: bool = False,
    min_labels: int = 2,
    max_labels: int = 10,
) -> list[str]:
    """From a specification given according to the :obj:`random_attribute_column_type`
    type, add attribute columns to the given dataframe and return the name of added
    columns.

    Depending on ``is_list``, it will be either an attribute column, where each row
    has a single value, taken from a fixed set of possible string labels or an attribute
    list column where each row has a subsset of values from a fixed superset of possible
    string labels.

    Args:
        input_dataframe: DataFrame which will be assigned new columns
        columns_specs: specification of columns, according to the aforementioned
            syntax
        numpy_generator: random generator for numpy arrays
        fake_generator: random generator for random unique words
        is_list: if set to True, will construct list attribute columns. Otherwise, will
            construct simple attribute columns. Defaults to False
        min_labels: When number of labels if not specified, minimum random number of
            labels to generate for the current column. Defaults to 2.
        max_labels: When number of labels if not specified, maximum random number of
            labels to generate for the current column. Defaults to 10.

    Returns:
        The header of added columns. Useful to keep track of list attribute columns
        to booleanize them.
    """

    def random_labels(n_labels: int) -> list[str]:
        return [fake_generator.unique.word() for _ in range(n_labels)]

    def is_float_sequence(
        sequence: Sequence[str] | Sequence[float],
    ) -> TypeGuard[Sequence[float]]:
        types = set(map(type, sequence))
        if types not in [{float}, {str}]:
            raise ValueError(
                "The input specification accepts sequence of only float or only string"
                f" labels for dictionary values, got {types} instead"
            )
        return isinstance(sequence[0], float)

    # This typeguard is needed because typeguard is mostly lacking the type-narrowing
    # feature. See PEP 742 https://peps.python.org/pep-0742/
    def is_str_sequence(
        sequence: Sequence[str] | Sequence[float],
    ) -> TypeGuard[Sequence[str]]:
        return isinstance(sequence[0], str)

    def construct_detailed_column_spec(
        input_spec: int | Sequence[str] | Sequence[float] | dict[str, float],
    ) -> tuple[Sequence[str], Sequence[float] | None]:
        if isinstance(input_spec, int):
            labels = random_labels(n_labels=input_spec)
            probs = None
        elif isinstance(input_spec, dict):
            labels = list(input_spec.keys())
            probs = list(input_spec.values())
        else:
            if is_float_sequence(input_spec):
                labels = random_labels(len(input_spec))
                probs = input_spec
            else:
                assert is_str_sequence(input_spec)
                labels = input_spec
                probs = None
        return labels, probs

    # specification dictionary: key is the name of the column header, and values
    # is two vectors: names and probabilities. probabilities vector can be None
    column_labels: dict[str, tuple[Sequence[str], Sequence[float] | None]] = {}
    if isinstance(columns_specs, int):
        for _ in range(columns_specs):
            n_labels = numpy_generator.integers(min_labels, max_labels)
            header_name = fake_generator.unique.word()
            labels = random_labels(n_labels)
            column_labels[header_name] = (labels, None)
    elif isinstance(columns_specs, dict):
        for header_name, specific_column_spec in columns_specs.items():
            column_labels[header_name] = construct_detailed_column_spec(
                specific_column_spec
            )
    else:  # Simple sequence
        for specific_column_spec in columns_specs:
            if isinstance(specific_column_spec, str):
                header_name = specific_column_spec
                specific_column_spec = numpy_generator.integers(min_labels, max_labels)
            else:
                header_name = fake_generator.unique.word()
            column_labels[header_name] = construct_detailed_column_spec(
                specific_column_spec
            )
    for header_name, (labels, probs) in column_labels.items():
        input_dataframe[header_name] = construct_attribute_column(
            numpy_generator, len(input_dataframe), labels, probs, is_list
        )

    return list(column_labels.keys())


def dummy_dataset(
    n_imgs: int = 2,
    n_annot: int = 2,
    n_labels: int = 3,
    split_names: None | str | Sequence[str] = ("train", "val", "eval"),
    split_shares: Sequence[float] = (0.8, 0.1, 0.1),
    n_list_columns_images: random_attribute_column_type = 0,
    n_list_columns_annotations: random_attribute_column_type = 0,
    n_attribute_columns_images: random_attribute_column_type = 0,
    n_attributes_columns_annotations: random_attribute_column_type = 0,
    booleanize: Literal["all", "random", "none"] = "none",
    keypoints_share: float = 0,
    add_confidence: bool = False,
    generate_real_images: bool = False,
    seed: int = 0,
    **existing_elements,
) -> Dataset:
    """Generate a Dummy dataset for demonstration purpose

    Might also be used for tests

    Args:
        n_imgs: number of frame in the fake dataset
        n_annot: number of annotations
        n_labels: length of the label map
        split_names: sequence containing names of the splits to apply to the dataset as
            a column of images dataframe. If set to None, no "split" column will be
            added to the images dataframe. If empty, will assume all splits are
            ``None``. If not empty, and with 2 elements or more, must be the same size
            as ``split_shares``. Defaults to ``("train", "val", "eval")``.
        split_shares: sequence containing share of each split whose name was given in
            ``split_names``. The ith element in ``split_shares`` represents the share
            (written as a float number between 0 and 1) of the dataset that will be
            assigned to this split. If ``split_names`` is empty or has a length of 1, it
            will be ignored. Otherwise, its size must match length of ``split_names``,
            and the value must all add up to 1. Defaults to ``(0.8, 0.1, 0.1)``.
        n_list_columns_images: Definition of the attribute lists columns for images.
            A list column cell contains a subset of a larger set of possible
            attributes, fixed for the whole columns, in the form of a list or a set.
            These columns are designed to be booleanized and are created with
            the function :func:`~construct_list_column`.
            See :obj:`random_attribute_column_type` for an in depth explanation of the
            syntax. Defaults to 0
        n_list_columns_annotations: number of list columns to add to the annotations
            dataframe. A list column cell contains a subset of a larger set of possible
            attributes, fixed for the whole columns, in the form of a list or a set.
            These columns are designed to be booleanized and are created with
            the function :func:`~construct_list_column`.
            See :obj:`random_attribute_column_type` for an in depth explanation of the
            syntax. Defaults to 0
        n_attribute_columns_images: number of attributes columns to add to the images
            dataframe. An attribute column cell contains one element for a set fixed for
            the whole column. These columns are created with the function
            :func:`~construct_list_column`. See :obj:`random_attribute_column_type`
            for an in depth explanation of the syntax. Defaults to 0
        n_attributes_columns_annotations: number of attributes columns to add to the
            annotations dataframe. An attribute column cell contains one element for a
            set fixed for the whole column. These columns are created with the function
            :func:`~construct_list_column`. See :obj:`random_attribute_column_type`
            for an in depth explanation of the syntax. Defaults to 0
        booleanize: how to booleanize the list columns. Can be "all", "random" and
            "none". Defaults to "none".

            - "all" means all the list columns will converted to multiple boolean
              columns
            - "none" means the list columns will be unchanged
            - "random" means a random number of list columns will be booleanized. The
              number of booleanized columns is chosen randomly, and the choice of
              these n booleanized columns is also done randomly.
        keypoints_share: Share of bounding box which are keypoints, i.e. with a height
            and width of 0. Set it to 1 to only have keypoints, and to 0 to have no
            keypoint. Defaults to 0.
        add_confidence: If set to True, will add a "confidence" column to annotations
            with random values between 0 and 1. Use this option to generate random
            predictions, to be used in e.g. an evaluator. Defaults to False.
        generate_real_images:
            if set to True, will generate random images and save them in the ``/tmp/``
            folder under a random file name. Otherwise, will just generate random
            file path to images without creating any. Defaults to False.
        seed: seed number for the generation. This will ensure that for a given seed
            number, the same dataset will be created.
        **existing_elements: optional existing dataset elements that you want not to be
            random.

    Returns:
        Dummy generated dataset

    Example:
        >>> dummy_dataset()
        Dataset object containing 2 images and 2 objects
        Name :
            inside_else_memory
        Images root :
            such/serious
        Images :
            width  height      relative_path   type  split
        id
        0     342     136       help/me.jpeg  .jpeg  train
        1     377     167  whatever/wait.png   .png  train
        Annotations :
            image_id category_str  category_id  ...  box_y_min   box_width  box_height
        id                                      ...
        0          0         step           15  ...  73.932999   71.552480   42.673983
        1          0          why           19  ...   4.567638  248.551257  122.602211
        <BLANKLINE>
        [2 rows x 8 columns]
        Label map :
        {15: 'step', 19: 'why', 25: 'interview'}

        Change the seed option to another random dataset following the same rules

        >>> dummy_dataset(seed=1)
        Dataset object containing 2 images and 2 objects
        Name :
            shake_effort_many
        Images root :
            care/suggest
        Images :
            width  height        relative_path  type  split
        id
        0     955     229  determine/story.jpg  .jpg  train
        1     131     840       air/method.bmp  .bmp  train
        Annotations :
            image_id category_str  category_id  ...   box_y_min   box_width  box_height
        id                                      ...
        0          1       listen           14  ...  276.974642    9.718823  184.684056
        1          0        reach           22  ...    6.311037  123.141689  174.239136
        <BLANKLINE>
        [2 rows x 8 columns]
        Label map :
        {14: 'listen', 15: 'marriage', 22: 'reach'}

        Use the ``split_share`` and ``split_names`` to set splits values.
        Use the ``keypoints_share`` option to set a share of bounding box with size of 0

        >>> dataset = dummy_dataset(
        ...     10,
        ...     100,
        ...     split_shares=(0.5, 0.5),
        ...     split_names=("foo", "bar"),
        ...     keypoints_share=0.3,
        ...     add_confidence=True,
        ... )
        >>> dataset
        Dataset object containing 10 images and 100 objects
        Name :
            inside_else_memory
        Images root :
            such/serious
        Images :
            width  height           relative_path   type split
        id
        0     342     645            help/me.jpeg  .jpeg   foo
        1     377     973       whatever/wait.png   .png   foo
        2     136     756        chair/mother.gif   .gif   bar
        3     167     669  someone/challenge.jpeg  .jpeg   foo
        4     114     589  successful/present.bmp   .bmp   bar
        5     257     603           no/where.jpeg  .jpeg   foo
        6     831     941          play/take.tiff  .tiff   foo
        7     684     349           bit/force.gif   .gif   bar
        8     921     834           way/back.tiff  .tiff   bar
        9     553     703      marriage/give.tiff  .tiff   foo
        Annotations :
            image_id category_str  category_id  ...   box_width  box_height  confidence
        id                                      ...
        0          0    interview           25  ...   11.569934  591.860047    0.136767
        1          3         step           15  ...   70.680613  101.235900    0.663684
        2          8    interview           25  ...    0.000000    0.000000    0.749956
        3          5          why           19  ...   99.047865  266.499060    0.163943
        4          0          why           19  ...   69.419403   61.451991    0.689302
        ..       ...          ...          ...  ...         ...         ...         ...
        95         7         step           15  ...  518.765436   55.277118    0.942361
        96         0         step           15  ...    0.000000    0.000000    0.802246
        97         5    interview           25  ...    0.000000    0.000000    0.122368
        98         4          why           19  ...   89.054816  254.947600    0.124429
        99         9          why           19  ...  181.630916   86.810354    0.616242
        <BLANKLINE>
        [100 rows x 9 columns]
        Label map :
        {15: 'step', 19: 'why', 25: 'interview'}
        >>> (dataset.annotations["box_width"] > 0).value_counts() / dataset.len_annot()
        box_width
        True     0.69
        False    0.31
        Name: count, dtype: float64

        Add list columns, that can be booleanized later

        >>> dummy_dataset(n_list_columns_images=1, n_list_columns_annotations=1)
        Dataset object containing 2 images and 2 objects
        Name :
            inside_else_memory
        Images root :
            such/serious
        Images :
            width  height  ...  split                            discover
        id                 ...
        0     342     136  ...  train                  [chair, challenge]
        1     377     167  ...  train  [someone, beyond, present, enough]
        <BLANKLINE>
        [2 rows x 6 columns]
        Annotations :
            image_id category_str  ...  box_height                                  where
        id                         ...
        0          0         step  ...   42.673983         [take, play, week, force, bit]
        1          0          why  ...  122.602211  [no, season, take, play, choice, bit]
        <BLANKLINE>
        [2 rows x 9 columns]
        Label map :
        {15: 'step', 19: 'why', 25: 'interview'}

        Or booleanize them right away

        >>> dummy_dataset(
        ...     n_list_columns_images=1, n_list_columns_annotations=1, booleanize="all"
        ... )
        Dataset object containing 2 images and 2 objects
        Name :
            inside_else_memory
        Images root :
            such/serious
        Images :
            width  height  ... discover.present discover.someone
        id                 ...
        0     342     136  ...            False            False
        1     377     167  ...             True             True
        <BLANKLINE>
        [2 rows x 11 columns]
        Annotations :
            image_id category_str  category_id  ... where.season  where.take  where.week
        id                                      ...
        0          0         step           15  ...        False        True        True
        1          0          why           19  ...         True        True       False
        <BLANKLINE>
        [2 rows x 16 columns]
        Label map :
        {15: 'step', 19: 'why', 25: 'interview'}

        Add attribute columns which then are transformed into categorical columns.

        >>> example = dummy_dataset(
        ...     n_attribute_columns_images={"a": 2, "b": 3},
        ...     n_list_columns_annotations=2,
        ... )
        >>> example
        Dataset object containing 2 images and 2 objects
        Name :
            inside_else_memory
        Images root :
            such/serious
        Images :
            width  height      relative_path   type  split     a      b
        id
        0     342     136       help/me.jpeg  .jpeg  train  play  force
        1     377     167  whatever/wait.png   .png  train  take  force
        Annotations :
            image_id  ...         where
        id            ...
        0          0  ...            []
        1          0  ...  [no, season]
        <BLANKLINE>
        [2 rows x 10 columns]
        Label map :
        {15: 'step', 19: 'why', 25: 'interview'}
        >>> example.images["b"]
        id
        0    force
        1    force
        Name: b, dtype: category
        Categories (3, object): ['week', 'choice', 'force']

        Instead of integers, use lists of probabilities to steer the distribution of
        attributes.

        >>> example = dummy_dataset(
        ...     200, n_attribute_columns_images=[[0.1, 0.1, 0.8]], seed=1
        ... )
        >>> example
        Dataset object containing 200 images and 2 objects
        Name :
            shake_effort_many
        Images root :
            care/suggest
        Images :
            width  height        relative_path   type  split could
        id
        0      955     488  determine/story.jpg   .jpg  train  note
        1      131     895       air/method.bmp   .bmp  train  firm
        2      229     880   political/lead.jpg   .jpg  train  firm
        3      840     384        like/safe.bmp   .bmp  train  note
        4      953     668      suffer/set.jpeg  .jpeg  train  note
        ..     ...     ...                  ...    ...    ...   ...
        195    122     437    state/almost.tiff  .tiff  train  firm
        196    752     300     weight/tend.jpeg  .jpeg  train  note
        197    554     228  remember/summer.png   .png  train  note
        198    688     605       yet/though.png   .png   eval  note
        199    243     227   describe/road.tiff  .tiff  train  note
        <BLANKLINE>
        [200 rows x 6 columns]
        Annotations :
            image_id category_str  category_id  ...   box_y_min   box_width  box_height
        id                                      ...
        0         77        reach           22  ...   45.427512   40.116677  318.073851
        1        137     marriage           15  ...  202.481384  435.389400  475.375279
        <BLANKLINE>
        [2 rows x 8 columns]
        Label map :
        {14: 'listen', 15: 'marriage', 22: 'reach'}
        >>> example.images["could"].value_counts() / len(example)
        could
        note    0.82
        firm    0.09
        lead    0.09
        Name: count, dtype: float64

        Finally, you can generate fake images as well if you want to test the io
        functions that need images to be valid.

        >>> dataset = dummy_dataset(generate_real_images=True)
        >>> dataset
        Dataset object containing 2 images and 2 objects
        Name :
            inside_else_memory
        Images root :
            /tmp/such/serious
        Images :
            width  height      relative_path   type  split
        id
        0     342     136       help/me.jpeg  .jpeg  train
        1     377     167  whatever/wait.png   .png  train
        Annotations :
            image_id category_str  category_id  ...  box_y_min   box_width  box_height
        id                                      ...
        0          0         step           15  ...  73.932999   71.552480   42.673983
        1          0          why           19  ...   4.567638  248.551257  122.602211
        <BLANKLINE>
        [2 rows x 8 columns]
        Label map :
        {15: 'step', 19: 'why', 25: 'interview'}
        >>> dataset.check()
        Checking Image and annotations Ids ...
        Checking Bounding boxes ..
        Checking label map ...
        Checking images are valid ...
    """
    gen = np.random.default_rng(seed)
    Faker.seed(seed=seed)
    fake_generator = Faker()
    images_root = existing_elements.get(
        "images_root", Path("/".join(fake_generator.words(2)))
    )
    if generate_real_images and not images_root.is_absolute():
        images_root = "/tmp" / images_root
    dataset_name = existing_elements.get(
        "dataset_name", "_".join(fake_generator.words(3))
    )

    if "label_map" in existing_elements:
        label_map = existing_elements["label_map"]
    else:
        label_ids = gen.integers(0, 10 * n_labels, size=n_labels)
        label_map = {
            int(label_id): fake_generator.unique.word() for label_id in label_ids
        }
    if "images" in existing_elements:
        images = existing_elements["images"]
        assert isinstance(images, pd.DataFrame), "Images can only be a dataframe"
        image_ids = images.index
        n_imgs = len(image_ids)
    else:
        image_ids = np.arange(n_imgs)
        image_paths = [
            Path(fake_generator.file_path(depth=1, category="image", absolute=False))
            for i in range(n_imgs)
        ]

        width = gen.integers(100, 1000, size=n_imgs)
        height = gen.integers(100, 1000, size=n_imgs)
        images = pd.DataFrame(
            data={
                "width": width,
                "height": height,
                "relative_path": image_paths,
            },
            index=image_ids,
        )

    annot_ids = np.arange(n_annot)
    annot_image_ids = gen.choice(image_ids, size=n_annot)

    category_ids = gen.choice(a=np.array(list(label_map.keys())), size=n_annot)
    bbox = gen.random((2, 2, n_annot))
    box_x_min, box_y_min = bbox.min(axis=0)
    box_width, box_height = np.abs(bbox[0] - bbox[1])
    if keypoints_share > 0:
        to_zero = gen.choice(
            a=[True, False], size=n_annot, p=[keypoints_share, 1 - keypoints_share]
        )
        box_width[to_zero] = 0
        box_height[to_zero] = 0

    bbox = np.stack([box_x_min, box_y_min, box_width, box_height], axis=1)

    if split_names is not None:
        if isinstance(split_names, str):
            split_names = [split_names]
        n_splits = len(split_names)
        if n_splits == 0:
            images["split"] = None
        elif n_splits == 1:
            images["split"] = split_names[0]
        else:
            if len(split_shares) != n_splits:
                raise ValueError(
                    "Size mismatch between 'split_names' and 'split_shares'"
                    f" ({len(split_names)} vs {len(split_shares)})"
                )
            if sum(split_shares) != 1:
                raise ValueError(
                    "Split share values must addup to 1. Got"
                    f" {sum(split_shares)} instead"
                )
            split = gen.choice(list(split_names), size=n_imgs, p=list(split_shares))
            images["split"] = split
    bbox = import_bbox(
        bbox, images_df=images, image_ids=annot_image_ids, input_format="xywh"
    )
    bbox.index = pd.Index(annot_ids)
    annotations = pd.DataFrame(
        data={
            "image_id": annot_image_ids,
            "category_id": category_ids,
        },
        index=annot_ids,
    )

    if add_confidence:
        annotations["confidence"] = gen.random(n_annot)

    annotations = pd.concat([annotations, bbox], axis="columns")

    columns_to_booleanize = []

    if n_list_columns_images:
        columns_to_booleanize.extend(
            set_attribute_columns_labels(
                input_dataframe=images,
                columns_specs=n_list_columns_images,
                numpy_generator=gen,
                fake_generator=fake_generator,
                is_list=True,
            )
        )

    if n_list_columns_annotations:
        columns_to_booleanize.extend(
            set_attribute_columns_labels(
                input_dataframe=annotations,
                columns_specs=n_list_columns_annotations,
                numpy_generator=gen,
                fake_generator=fake_generator,
                is_list=True,
            )
        )

    if n_attribute_columns_images:
        set_attribute_columns_labels(
            input_dataframe=images,
            columns_specs=n_attribute_columns_images,
            numpy_generator=gen,
            fake_generator=fake_generator,
            is_list=False,
        )

    if n_attributes_columns_annotations:
        set_attribute_columns_labels(
            input_dataframe=annotations,
            columns_specs=n_attributes_columns_annotations,
            numpy_generator=gen,
            fake_generator=fake_generator,
        )

    dataset = Dataset(
        label_map=label_map,
        images=images,
        annotations=annotations,
        images_root=images_root,
        dataset_name=dataset_name,
    )

    if booleanize == "all":
        dataset = dataset.booleanize(columns_to_booleanize)
    elif booleanize == "random":
        subset_to_booleanize = gen.choice(
            columns_to_booleanize,
            size=gen.integers(len(columns_to_booleanize)),
            replace=False,
        )
        dataset = dataset.booleanize(subset_to_booleanize)
    elif booleanize != "none":
        raise ValueError(
            "Invalid booleanize option. Possible values are 'all', 'random' or 'none',"
            f" got {booleanize}"
        )
    if generate_real_images:
        for i, image_data in dataset.images.iterrows():
            image_path = dataset.images_root / image_data["relative_path"]
            format = image_data["type"][1:]
            if format.lower() == "jpg":
                format = "jpeg"
            image_path.parent.mkdir(parents=True, exist_ok=True)
            with open(image_path, "wb") as f:
                image_data_stream = fake_generator.image(
                    image_format=format,
                    size=(image_data["width"], image_data["height"]),
                )
                f.write(image_data_stream)
    return dataset



"""Set of utility module to manipulate dataset objects."""

import types

from .bbox_converter import column_names_from_format_string

""" Conventions for Lours """
BBOX_COLUMN_NAMES = column_names_from_format_string("XYWH")
DISPLAY_NESTED_COLUMNS = False
DISPLAY_UNBOOLEANIZED = False


def try_import_fiftyone() -> types.ModuleType:
    """Try to import fiftyone and display an informative error if import was
    unsuccessful.

    Raises:
        ImportError: raise an error if the module could not be imported. It can be
            either because it's not installed, or because the database package
            (fyftone-db) is not compatible with the current distribution.

    Returns:
        types.ModuleType: the imported fiftyone module
    """
    try:
        import fiftyone as fo
    except ImportError as e:
        raise ImportError(
            "Fiftyone could not be loaded, make sure you have installed Lours with the"
            " 'fiftyone' extra."
        ) from e
    except Exception as e:
        raise ImportError(
            "Fiftyone is installed but cannot be loaded, you might have to install a"
            " custom 'fyftone-db' for your OS. See"
            " https://docs.voxel51.com/getting_started/troubleshooting.html#alternative-linux-builds"
        ) from e
    return fo



from os.path import commonpath
from pathlib import Path

import pandas as pd

from lours.dataset import Dataset
from lours.utils.column_booleanizer import broadcast_booleanization
from lours.utils.label_map_merger import merge_label_maps
from lours.utils.testing import assert_frame_intersections_equal


def merge_datasets(
    dataset1: Dataset,
    dataset2: Dataset,
    allow_overlapping_image_ids: bool = True,
    realign_label_map: bool = False,
    ignore_index: bool = False,
    mark_origin: bool = False,
    overwrite_origin: bool = False,
) -> Dataset:
    """Merge two datasets and return a unique dataset object containing
    Samples from both. Result's images_root will be the common path of both
    datasets, and the image relative paths will be updated accordingly.
    Result's label map will be the superset of both label map,
    provided one is included in the other.

    Notes:
        - If possible, booleanized columns for images and annotations will be broadcast
          together. See :func:`lours.utils.column_booleanizer.broadcast_booleanization`
        - If one of the dataset has an absolute path as ``images_root``, the other
          dataset images root path will also be converted to absolute.
        - If both datasets have the same name, the output will have the same name
          as well.
        - If datasets have a different name, the output will have the concatenation
          of both names separate by a "+" sign. The merge output of "A" and "B" will
          be thus names "A+B".
        - If one dataset has no name (``dataset.name`` is ``None``), the output will
          take the name of the other.
        - If ``mark_origin`` is selected, it will be effective only if datasets have
          different actual names (not ``None``)

    Args:
        dataset1: First dataset to merge.
        dataset2: Second dataset to merge with dataset1. This dataset must be
            compatible with the first one, i.e. one label map is included with the
            other,  image and annotation ids are mutually exclusives between
            datasets (unless `ignore_index` is False), and booleanized columns are
            compatible with each other.
        allow_overlapping_image_ids: if set to True, will try to join images
            dataframes with overlapping ids. The whole rows (i.e. with values from
            columns present in both dataframes) must match, as well as
            the images_root. In that case, annotations with this image_id
            (from self or other) will be assumed to come from the same image.
            Defaults to True
        realign_label_map: If set to True, will try to remap classes of dataset2 to
            match the label map fo dataset1, to avoid a potential error due to
            incompatible label maps.
        ignore_index: if set to True, will ignore overlapping ids
            for images and annotations and reset them. Will update the ``image_id``
            column in the annotations accordingly. Note that this option makes the
            former option useless. Defaults to False.
        mark_origin: If set to True, and if both datasets have a different name, will
            add two columns "origina_dataset_name" and "origin" for images and
            annotations dataframes, indicating respectively the name
            of the origin dataset, and its id in the original dataset. Defaults to True.
        overwrite_origin: If set to True, will overwrite already existing columns in
            input datasets dataframes. Otherwise, will only mark origin if it's not
            present. Defaults to False.

    Raises:
        ValueError: Error if the two datasets are incompatible (see above)

    Returns:
        Merged dataset.
    """
    needs_new_name = (
        dataset1.dataset_name is not None
        and dataset2.dataset_name is not None
        and dataset2.dataset_name != dataset1.dataset_name
    )

    def mark_origin_to_dataset(dataset: Dataset) -> Dataset:
        dataset_columns = set(dataset.images.columns) & set(dataset.annotations.columns)
        if (
            "origin" in dataset_columns
            and "origin_id" in dataset_columns
            and not overwrite_origin
        ):
            return dataset
        return dataset.from_template(
            images=dataset.images.assign(
                origin=dataset.dataset_name,
                origin_id=dataset.images.index,
            ),
            annotations=dataset.annotations.assign(
                origin=dataset.dataset_name,
                origin_id=dataset.annotations.index,
            ),
        )

    if mark_origin and needs_new_name:
        dataset1 = mark_origin_to_dataset(dataset1)
        dataset2 = mark_origin_to_dataset(dataset2)

    if ignore_index:
        return merge_datasets(
            dataset1.reset_index(),
            dataset2.reset_index(len(dataset1)),
            allow_overlapping_image_ids=True,
            ignore_index=False,
        )

    # images_root to grab images might not be the same
    # Get the common path and update images relative paths to
    # be relative to that new path
    if dataset1.images_root != dataset2.images_root:
        if dataset1.images_root.is_absolute() or dataset2.images_root.is_absolute():
            images_root = commonpath(
                [
                    dataset1.images_root.absolute(),
                    dataset2.images_root.absolute(),
                ]
            )
        else:
            images_root = commonpath([dataset1.images_root, dataset2.images_root])
        images_root = Path(images_root)
        dataset1 = dataset1.reset_images_root(images_root)
        dataset2 = dataset2.reset_images_root(images_root)
    else:
        images_root = dataset1.images_root

    if realign_label_map:
        dataset2 = dataset2.remap_from_other(dataset1)
        label_map = merge_label_maps(
            dataset1.label_map, dataset2.label_map, method="outer"
        )
    else:
        label_map = merge_label_maps(
            dataset1.label_map, dataset2.label_map, method="outer"
        )

    dataset1_images, dataset2_images, booleanized_image_columns = (
        broadcast_booleanization(
            dataset1.images,
            dataset2.images,
            booleanized_columns1=dataset1.booleanized_columns["images"],
            booleanized_columns2=dataset2.booleanized_columns["images"],
        )
    )
    dataset1_annotations, dataset2_annotations, booleanized_annotations_columns = (
        broadcast_booleanization(
            dataset1.annotations,
            dataset2.annotations,
            booleanized_columns1=dataset1.booleanized_columns["annotations"],
            booleanized_columns2=dataset2.booleanized_columns["annotations"],
        )
    )

    dataset1_images_ids = set(dataset1.images.index)
    dataset2_images_ids = set(dataset2.images.index)
    mutual_images_ids = dataset1_images_ids & dataset2_images_ids
    dataset1_images_columns = set(dataset1.images.columns)
    dataset2_images_columns = set(dataset2.images.columns)
    mutual_images_columns = dataset1_images_columns & dataset2_images_columns

    if mutual_images_ids and not allow_overlapping_image_ids:
        raise ValueError(
            "Overlapping image ids not permitted. Consider using the"
            " allow_overlapping_image_ids or ignore_index options"
        )

    assert_frame_intersections_equal(
        dataset1_images.drop(["origin", "origin_id"], axis=1, errors="ignore"),
        dataset2_images.drop(["origin", "origin_id"], axis=1, errors="ignore"),
    )

    # Concat horizontally by extending images from dataset1 with columns from dataset2
    # and then vertically by extending images with dataset2 images which id is not
    # in dataset1 images index.
    dataset1_images = dataset1_images.join(
        dataset2_images.loc[
            list(mutual_images_ids),
            list(dataset2_images_columns - mutual_images_columns),
        ]
    )
    dataset2_images = dataset2_images.loc[
        list(dataset2_images_ids - mutual_images_ids), :
    ]

    images = pd.concat([dataset1_images, dataset2_images])

    # Merge annotations.
    mutual_instance_ids = set(dataset1_annotations.index).intersection(
        set(dataset2_annotations.index)
    )
    # Only reset index of dataset2's annotations if there is overlap.
    # However, keep the index of first dataset as is
    if mutual_instance_ids:
        dataset2_annotations.index += (
            dataset1_annotations.index.max() - dataset2_annotations.index.min() + 1
        )
    annotations = pd.concat([dataset1_annotations, dataset2_annotations])
    annotations.index.name = "id"

    if needs_new_name:
        output_dataset_name = f"{dataset1.dataset_name}+{dataset2.dataset_name}"
    elif dataset2.dataset_name is None:
        output_dataset_name = dataset1.dataset_name
    else:
        output_dataset_name = dataset2.dataset_name
    output = Dataset(
        images_root=images_root,
        images=images,
        annotations=annotations,
        label_map=label_map,
        dataset_name=output_dataset_name,
    )
    output.booleanized_columns["images"] = booleanized_image_columns
    output.booleanized_columns["annotations"] = booleanized_annotations_columns

    return output



from collections.abc import Iterable
from functools import partial

import pandas as pd


def booleanize(
    input_df: pd.DataFrame,
    column_names: str | Iterable[str] | None = None,
    separator: str = ".",
    **possible_values: set | None,
) -> pd.DataFrame:
    """Convert given column in input DataFrame from lists to boolean

    This is mainly used when a particular attribute can have multiple possible
    values at once.

    Every possible value given will be tested to see if it's inside every row's list
    which will give a boolean column.

    In the end, the column will be dropped and N new boolean columns will be created
    with the name in form ``{column_name}{separator}{value}``

    Args:
        input_df: DataFrame on which performing the booleanization. The operation is not
            inplace.
        column_names: columns to convert. After conversion, it will be dropped
            from input DataFrame. Can be either a single string or a list of strings.
        separator: character used to separate original column and value. Defaults to '.'
        **possible_values: kwargs for sets of possible values. Each key in this
            dictionary must match a column name. If the corresponding value is None,
            will deduce it from all occurrences in lists of column given by key.
            Defaults to None.

    Raises:
        KeyError: The given ``column_name`` must be in the columns of ``input_df``
        TypeError: When for a particular column possible values need to be deduced,
            the column must have value that are all iterable except strings.

    Returns:
        New dataset with multiple boolean columns in the form
        ``{column_name}{separator}{value}``.
    """

    def is_true(iterable, value) -> bool:
        try:
            return value in iterable
        except TypeError:
            return value == iterable

    if column_names is None:
        column_names = set()
    if isinstance(column_names, str):
        column_names = {column_names}
    else:
        column_names = {*column_names}
    if possible_values:
        column_names = set(column_names).union(possible_values.keys())
    elif not column_names:
        # Nothing to booleanize, return immediately
        return input_df

    for column_name in column_names:
        if input_df[column_name].dropna().apply(lambda x: isinstance(x, str)).any():
            raise TypeError(
                f"Column {column_names} cannot contain a single string, use lists"
                " instead"
            )
        enum = possible_values.get(column_name, None)
        if enum is None:
            enum = set().union(*input_df[column_name].dropna().to_list())
        column_name_index = input_df.columns.get_loc(column_name)
        before_columns = input_df.columns[:column_name_index]
        after_columns = input_df.columns[column_name_index + 1 :]
        new_columns = []
        for v in enum:
            booleanized_column_name = f"{column_name}{separator}{v}"
            input_df = input_df.assign(
                **{
                    booleanized_column_name: input_df[column_name].apply(
                        partial(is_true, value=v)
                    )
                }
            )
            new_columns.append(booleanized_column_name)

        input_df = input_df[
            [
                *before_columns,
                *new_columns,
                *after_columns,
            ]
        ]
    return input_df


def broadcast_booleanization(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    booleanized_columns1: Iterable[str] = (),
    booleanized_columns2: Iterable[str] = (),
    ignore_index: bool = False,
    separator: str = ".",
) -> tuple[pd.DataFrame, pd.DataFrame, set[str]]:
    """Broadcast two dataframes so that they have the same booleanized columns.

    Booleanized columns from ``df1`` that are not present in ``df2`` will be created and
    set to False and vice versa.

    Note: if ``ignore_index`` is set to False, the overlapping ids will be set to the
    value in the other dataframe instead of just False


    Args:
        df1: first dataframe to broadcast
        df2: second dataframe to broadcast
        booleanized_columns1: Columns in ``df1`` that are booleanized. Defaults to ().
        booleanized_columns2: Columns in ``df2`` that are booleanized. Defaults to ().
        ignore_index: if set to True, will create boolean columns full of False
            regardless of index overlap between the two dataframes.
            If set to False, tries to retrieve boolean value in one dataframe
            from the other when creating the column. Defaults to False.
        separator: Character used to separate column prefix and value.
            Defaults to ".".

    Returns:
        tuple containing updated dataframes ``df1`` and ``df2`` with the same
        booleanized columns
    """
    booleanized_columns = set().union(booleanized_columns1, booleanized_columns2)
    for column in booleanized_columns:
        if column not in booleanized_columns2 and column in df2.columns:
            df2 = booleanize(df2, column, separator=separator)
        if column not in booleanized_columns1 and column in df1.columns:
            df1 = booleanize(df1, column, separator=separator)
        bool_columns1 = get_bool_columns(df1, column, separator)
        bool_columns2 = get_bool_columns(df2, column, separator)
        for bool_column in set().union(bool_columns1, bool_columns2):
            if bool_column not in df1.columns:
                df1 = df1.assign(
                    **{
                        bool_column: (
                            False
                            if ignore_index
                            else df2[bool_column].reindex(df1.index, fill_value=False)
                        )
                    }
                )
            if bool_column not in df2.columns:
                df2 = df2.assign(
                    **{
                        bool_column: (
                            False
                            if ignore_index
                            else df1[bool_column].reindex(df2.index, fill_value=False)
                        )
                    }
                )
    return df1, df2, booleanized_columns


def get_bool_columns(
    input_df: pd.DataFrame, column_prefix: str, separator: str = "."
) -> list[str]:
    """Given a prefix and a separator, get all columns that start with
    ``{column_prefix}{separator}``

    This is used in e.g. :func:`.debooleanize`

    Args:
        input_df: DataFrame to get the columns from
        column_prefix: Name of column prefix to retrieve boolean columns.
        separator: Character used to separate column prefix and value.
            Defaults to ".".

    Raises:
        ValueError: Raised when column following the pattern are not boolean

    Returns:
        List of columns that follow the pattern and will be used to construct the list.
    """
    full_prefix = f"{column_prefix}{separator}"

    columns = [
        name
        for name in input_df.columns
        if isinstance(name, str) and name.startswith(full_prefix)
    ]

    column_dtypes = input_df[columns].dtypes

    if any(
        dtype not in [bool, pd.BooleanDtype()]
        for column_name, dtype in column_dtypes.items()
    ):
        raise ValueError(
            f"Expected bool type for columns starting with {column_prefix}, but got"
            f" the following dtypes : {column_dtypes}."
        )
    return columns


def debooleanize(
    input_df: pd.DataFrame,
    column_prefixes: str | Iterable[str],
    separator: str = ".",
) -> pd.DataFrame:
    """Inverse operation of :func:`.booleanize`. Take all columns that start with
    ``{column_prefix}{separator}`` and, assuming they are all boolean columns, convert
    them into a single column of list values.

    Note:
        The column order will be preserved, the debooleanized column will be inserted
        at the same spot the multiple booleanized columns were.

    Args:
        input_df: Input DataFrame we will take the columns from.
        column_prefixes: Name of column prefix (or prefixes) to retrieve boolean
            columns. Also, the name of resulting column (or columns)
        separator: Character used to separate column prefix and value.
            Defaults to ".".

    Raises:
        TypeError: all columns with given prefix must be of boolean dtype

    Returns:
        pd.DataFrame: Resulting DataFrame, with all boolean column which name correspond
            to the prefix drop and a single column added with lists
    """
    if isinstance(column_prefixes, str):
        column_prefixes = [column_prefixes]
    for column_prefix in column_prefixes:
        full_prefix = f"{column_prefix}{separator}"

        columns = get_bool_columns(input_df, column_prefix, separator)
        if not columns:
            continue
        first_column_pos = input_df.columns.get_loc(columns[0])
        columns_df = input_df[columns]
        input_df = input_df.drop(columns=columns)
        columns_df = columns_df.rename(
            columns={n: n.replace(full_prefix, "") for n in columns_df.columns}
        )
        single_column = columns_df.apply(lambda x: x[x].index.tolist(), axis=1)
        input_df = input_df.assign(**{column_prefix: single_column})
        # Now reorder columns so that the newly created column is where the
        # booleanized ones were
        input_df = input_df[
            [
                *input_df.columns[:first_column_pos],
                column_prefix,
                *input_df.columns[first_column_pos:-1],
            ]
        ]
    return input_df



"""Registry for known useful preset.

to add a preset, add a new entry in remap_presets
value should be a tuple of two dictionaries,
with a ``old_id -> new_id`` mapping and a ``new_id -> new_name`` mapping
"""

from importlib.resources import files

import pandas as pd

__all__ = ["presets", "list_available_presets"]

presets = {}
presets_folder = files("lours") / "dataset" / "remap_presets"


for p in presets_folder.iterdir():
    if p.is_file() and p.name.endswith(".csv"):
        try:
            input_dataset, output_dataset = p.name.removesuffix(".csv").split("_to_")
        except ValueError as e:
            raise NameError(
                "Badly named csv preset file. Should be in the form "
                f"'<dataset1>_to_<dataset2>.csv', but got {p.name} instead."
            ) from e
        preset_df = pd.read_csv(p.open())
        indexed_preset = preset_df.set_index("input_category_id")
        preset_dict = indexed_preset["output_category_id"].to_dict()
        preset_names = (
            indexed_preset.groupby("output_category_id")["output_category_name"]
            .first()
            .to_dict()
        )
        presets[(input_dataset, output_dataset)] = (preset_dict, preset_names)
        is_invertible = preset_df["output_category_id"].is_unique
        if is_invertible and (output_dataset, input_dataset) not in presets:
            inverted_preset = preset_df.set_index("output_category_id")
            inverted_preset_dict = inverted_preset["input_category_id"].to_dict()
            inverted_preset_names = (
                inverted_preset.groupby("input_category_id")["input_category_name"]
                .first()
                .to_dict()
            )
            presets[(output_dataset, input_dataset)] = (
                inverted_preset_dict,
                inverted_preset_names,
            )


def list_available_presets():
    mapping_names = [f"{in_map}\t->\t{out_map}" for in_map, out_map in presets]
    return "\n".join(mapping_names)



"""Set of utility function to use json schemas for loading caipy json files

See Also:
    :ref:`Related tutorial </notebooks/6_demo_schemas.ipynb>`
"""

import json
import re
from functools import lru_cache
from importlib.resources import files
from pathlib import Path

import pandas as pd
import requests


def load_json_schema(schema_path: str | Path) -> dict:
    """Load JSON schema file, either from a url or a file path.

    If no schema path or url is given, an example following coco is loaded.

    Args:
        schema_path: Name of internal schema, or path to custom schema.

    Raises:
        KeyError: Errors when a string is given but no corresponding json file is found
            in the ``schemas`` folder.

    Returns:
        Loaded schema dictionary
    """
    if schema_path == "default":
        with (
            files("lours") / "dataset" / "io" / "schema_util" / "default-schema.json"
        ).open() as f:
            return json.load(f)
    if isinstance(schema_path, str) and (
        schema_path.startswith("https://") or schema_path.startswith("http://")
    ):
        response = requests.get(schema_path)
        response.raise_for_status()
        return response.json()
    with open(schema_path) as f:
        return json.load(f)


def get_enums(
    schema: dict, separator: str = ".", ignore_pattern: str = "a^"
) -> dict[str, set]:
    """From a schema, get column names that can be converted to sets of boolean columns.

    Each outputted column will be associated to the list of possible values in output
    dictionary

    Args:
        schema: JSON schema dict describing the expected format of input data
        separator: Separator to apply for path to get flattened paths in the dataset's
            DataFrames. Defaults to ".".
        ignore_pattern: column following this regex pattern will be ignored.
            Defaults to "a^".

    Returns:
        Dictionary describing enum columns and possible values
        (and thus created columns)
    """
    enums = {}
    for name, prop in schema["properties"].items():
        if re.match(ignore_pattern, name) or "type" not in prop:
            continue
        if (
            prop["type"] == "array"
            and "enum" in prop.get("items", {})
            and prop.get("uniqueItems", False)
        ):
            enums[name] = set(prop["items"]["enum"])
        elif prop["type"] == "object":
            sub_enums = get_enums(
                prop, separator=separator, ignore_pattern=ignore_pattern
            )
            for subname, values in sub_enums.items():
                enums[f"{name}{separator}{subname}"] = values
    return enums


def flatten_schema(
    schema: dict, separator: str = ".", prefix: str | None = None
) -> list[str]:
    """From a particular schema, get a list of expected key values if the schema was to
    be flattened by e.g. the function :func:`pandas.json_normalize`

    Note:
        This function is meant to be called recursively, hence the ``prefix`` option.

    Args:
        schema: JSON schema describing expected output format
        separator: Character used to separate name in flattened key. Defaults to ".".
        prefix: Prefix to apply to column names in output dictionary values.
            Defaults to None.

    Returns:
        list of flattened column names.
    """
    keys = []
    for name, prop in schema["properties"].items():
        normalized_name = name if prefix is None else separator.join((prefix, name))
        if prop.get("type") == "object":
            sub_tree = flatten_schema(
                schema=prop, separator=separator, prefix=normalized_name
            )
            keys.extend(sub_tree)
        else:
            keys.append(normalized_name)
    return keys


def get_dtypes_and_default_values(
    schema: dict, separator: str = "."
) -> tuple[dict, dict]:
    """Given a schema, find default values and dtypes to set to a flattened version of
    a dict corresponding to the schema.

    For optional integers and booleans we use pandas' Nullable dtypes when ``np.nan``
    is replaced with ``pd.NA``. Otherwise, these columns will get casted to float as
    soon as a value is missing. See :class:`pandas.BooleanDtype`
    and :class:`pandas.UInt64Dtype`

    Args:
        schema: JSON schema describing expected input format of input dicts.
        separator: Character used to separate name in flattened key. Defaults to ".".

    Returns:
        Dictionary with same keys as the flattened dictionary, and with the default
        values as values. If no default could be found (ambiguous type), the key is not
        present.
    """
    default_values = {}
    dtypes = {}
    dtype_mapping = {
        "integer": {True: "Int64", False: int},
        "bool": {True: "boolean", False: bool},
    }
    flattened_keys = flatten_schema(schema, separator=separator)
    for key in flattened_keys:
        schema_object = schema
        optional = False
        for part in key.split(separator):
            if part not in schema_object.get("required", []):
                optional = True
            schema_object = schema_object["properties"][part]
        default_value = schema_object.get("default")
        key_type = schema_object.get("type", "unknown")
        if optional:
            if default_value is not None:
                default_values[key] = default_value
            elif key_type == "array":
                default_values[key] = []
            elif key_type in ["integer", "bool"]:
                default_values[key] = pd.NA

        if key_type in dtype_mapping:
            dtypes[key] = dtype_mapping[key_type][optional]

    return default_values, dtypes


def fill_with_dtypes_and_default_value(
    schema: dict, input_dataframe: pd.DataFrame, separator: str = "."
) -> pd.DataFrame:
    """Given a schema and dataframe constructed on a list of corresponding dicts,
    avoid having NaN values by setting the default value when possible.

    It is expected that the DataFrame is constructed with :func:`pandas.json_normalize`

    Args:
        schema: JSON schema describing expected input format of input dicts.
        input_dataframe: input dataframe with possible missing values
            (and thus set to NaN)
        separator: Character used to separate name in flattened key. Defaults to ".".

    Returns:
        DataFrame similar to input_dataframe but with NaN replaced with default values
            when possible
    """
    default_values, dtypes = get_dtypes_and_default_values(schema, separator)
    for k, v in default_values.items():
        if k not in input_dataframe.columns:
            continue
        if isinstance(v, list):
            # Note that we don't use fillna here because it does not work with the
            # default value of []
            # See https://stackoverflow.com/questions/33199193/how-to-fill-dataframe-nan-values-with-empty-list-in-pandas
            isnull = input_dataframe[k].isna()
            input_dataframe.loc[isnull, k] = pd.Series([v] * isnull.sum()).values
        else:
            with pd.option_context("future.no_silent_downcasting", True):
                input_dataframe[k] = (
                    input_dataframe[k].fillna(v).infer_objects(copy=False)
                )
    dtypes_to_apply = {
        col: dtype for col, dtype in dtypes.items() if col in input_dataframe.columns
    }
    return input_dataframe.astype(dtypes_to_apply)


def get_remapping_dict_from_schema(
    schema: dict, separator: str = ".", prefix: str | None = None
) -> dict:
    """From a particular schema, get a nested dictionary similar to the expected format
    of given schema.

    Each value of that dictionary will be the name of column to get the value from in
    flattened DataFrame.

    Note:
        This function is meant to b called recursively, hence the ``prefix`` option.

    Args:
        schema: JSON schema describing expected output format
        separator: Character used to separate name in flattened key. Defaults to ".".
        prefix: Prefix to apply to column names in output dictionary values.
            Defaults to None.

    Returns:
        Nested dictionary following format described in schema, and providing mapping
        for nested DataFrames with flattened column names.
    """
    mapping_tree = {}
    for name, prop in schema["properties"].items():
        normalized_name = name if prefix is None else separator.join((prefix, name))
        if prop.get("type") == "object":
            sub_tree = get_remapping_dict_from_schema(
                schema=prop, separator=separator, prefix=normalized_name
            )
            mapping_tree[name] = sub_tree
        else:
            mapping_tree[name] = normalized_name
    return mapping_tree


@lru_cache
def get_remapping_dict_from_names(
    names: frozenset[str] | tuple[str, ...], separator: str = "."
) -> dict[str, list[str]]:
    """From a set of names, get the expected nested dictionary shape, assuming that
    a key with two names separated with the given separator means a nested dictionary
    shape.

    For example "a.b" means output shape is of the form ``{a: {b: value}}``

    Note:
        For the LRU cache to be used, the given names must hashable, either tuple or
        frozenset.

    Args:
        names: Set of names to parse the underlying structure from.
        separator: Character used to separate name in flattened key. Defaults to ".".

    Returns:
        Nested remapping dictionary with values set to flattened dictionary key to
        take values from.
    """
    output = {}
    for name in names:
        keys = name.split(separator)
        current = output
        for i, k in enumerate(keys):
            if i == len(keys) - 1:
                if k in current:
                    raise ValueError(
                        f"Cannot assign two values to key {name}."
                        f" Remapping dict state : {output}"
                    )
                current[k] = name
            else:
                if k not in current.keys():
                    current[k] = {}
                elif not isinstance(current[k], dict):
                    raise ValueError(
                        f"Cannot assign both a value and a dict to key {name}."
                        f" Remapping dict state : {output}"
                    )
                current = current[k]
    return output


def remap_dict(flattened_dict: dict, mapping_tree: dict | None = None) -> dict:
    """From a mapping tree, convert a flattened dict, possibly taken from a DataFrame
    into a nested dictionary.

    Args:
        flattened_dict: dictionary without sub-dictionary, easily readable by pandas.
        mapping_tree: nested dictionary following expected output shape.
            Each value represents. the key name from flattened dictionary to take the
            value from. If set to None, will deduce it from the key names and separator
            character ".". Defaults to None.

    Returns:
        Remapped nested dictionary
    """
    output_dict = {}
    if mapping_tree is None:
        mapping_tree = get_remapping_dict_from_names(frozenset(flattened_dict.keys()))
    for k, v in mapping_tree.items():
        if isinstance(v, dict):
            output_dict[k] = remap_dict(flattened_dict, v)
        else:
            output_value = flattened_dict.get(v, None)
            # Remove both empty lists and NaN/None values
            if isinstance(output_value, list):
                if not output_value:
                    continue
            else:
                isna = pd.isna(output_value)
                try:
                    if isna:
                        continue
                except ValueError:
                    # pd.isna either outputs a bool or a bool array when the input is
                    # iterable. In that case this raises a ValueError
                    # (Ambiguous truth value), which we ignore because then the object
                    # is clearly not na
                    if isna.any():  # pyright: ignore
                        raise ValueError(f"value contains nan : {output_value}")
                    pass
            output_dict[k] = output_value
    return output_dict



import json
from pathlib import Path

import pandas as pd
from tqdm.auto import tqdm

from ...utils import BBOX_COLUMN_NAMES
from ..dataset import Dataset
from .common import get_image_info, parse_annotation_name

BOX_XMIN, BOX_WIDTH, BOX_YMIN, BOX_HEIGHT = BBOX_COLUMN_NAMES


def from_crowd_human(
    annotation_odgt: Path | str,
    images_root: Path | str | None = None,
    visible_box: bool = True,
    dataset_name: str | None = None,
    split: str | None = None,
) -> Dataset:
    """Read a dataset in the format described for CrowdHuman

    See https://www.crowdhuman.org/download.html

    Args:
        annotation_odgt: annotation file using the format described in link given above.
        images_root: Folder where images are stored. Note that since annotations IDs are
            file names, all images must be at the root of this folder.
            If set to None, will use the image folder in
            the annotation file's parent folder. Defaults to None.
        visible_box: If set to True, will only take the visible bounding box. Otherwise,
            will take the whole bounding box, with possibly occluded parts.
        dataset_name: If specified, will be the dataset name, used when showing the
            dataset or exporting in other formats such as fiftyone. If not specified,
            the dataset name will be deduced from the name of the json file.
        split: Split value to give to the dataset. If set to None, will try to deduce
            it from the annotation file name. Defaults to None.

    Raises:
        FileNotFoundError: Will raise an error if the images_root folder does not exist
            Unlike COCO or Caipy, image size information is not stored in annotations
            and thus need to be computed by loading picture's headers.

    Returns:
        Dataset
    """
    annotation_odgt = Path(annotation_odgt)
    if images_root is None:
        images_root = annotation_odgt.parent / "images"
    else:
        images_root = Path(images_root)

    if not images_root.is_dir():
        raise FileNotFoundError(
            "Image folder needs to exist as CrowdHuman's annotations don't have image"
            " size information"
        )

    parsed_dataset_name, parsed_split = parse_annotation_name(
        annotations_file_path=annotation_odgt
    )
    if dataset_name is None:
        dataset_name = parsed_dataset_name
    if split is None:
        split = parsed_split

    with open(annotation_odgt) as f:
        lines = f.read().strip().split("\n")

    annots = [json.loads(line) for line in lines]
    image_dicts = []
    annotation_dicts = []
    label_map = {0: "person", 1: "head"}
    ids_map = {v: k for k, v in label_map.items()}
    for image_id, image_data in enumerate(tqdm(annots)):
        relative_path = Path(f"{image_data['ID']}.jpg")
        image_path = images_root / relative_path
        image_info = get_image_info(image_id, relative_path, image_path)
        image_dicts.append(image_info)

        current_annotations = []
        if "gtboxes" not in image_data:
            continue
        for annot in image_data["gtboxes"]:
            if annot["tag"] == "mask":
                continue
            if "ignore" in annot["extra"] and annot["extra"]["ignore"] == 1:
                continue

            *_, box_w, box_h = annot["fbox"]
            area = box_w * box_h
            *_, vbox_w, vbox_h = annot["vbox"]
            visible_share = vbox_w * vbox_h / area
            person_annotation = {
                **annot["extra"],
                "image_id": image_id,
                "category_id": ids_map["person"],
                "bbox": annot["vbox"] if visible_box else annot["fbox"],
                "visible_share": visible_share,
            }
            current_annotations.append(person_annotation)
            if "ignore" in annot["head_attr"] and annot["head_attr"]["ignore"] == 1:
                continue
            # We don't know how visible each head is
            # Sometimes it's occluded, sometimes it's even out of the screen
            head_annotation = {
                **annot["extra"],
                **annot["head_attr"],
                "image_id": image_id,
                "category_id": ids_map["head"],
                "bbox": annot["hbox"],
                "visible_share": float("nan"),
            }
            current_annotations.append(head_annotation)
        annotation_dicts.extend(current_annotations)
    images = pd.DataFrame.from_records(image_dicts).set_index("id")
    images["split"] = split

    annotations = pd.DataFrame.from_records(annotation_dicts)
    annotations.index.name = "id"
    bboxes = pd.DataFrame(list(annotations["bbox"]), index=annotations.index)
    for i, name in enumerate(BBOX_COLUMN_NAMES):
        annotations[name] = bboxes[i]
    annotations["split"] = split
    annotations = annotations.drop(["bbox", "ignore"], axis=1, errors="ignore")
    annotations["unsure"] = annotations["unsure"].fillna(0)

    return Dataset(
        images_root=images_root,
        images=images,
        annotations=annotations,
        label_map=label_map,
        dataset_name=dataset_name,
    ).cap_bounding_box_coordinates()



"""Set of functions and Classes dedicated for object detection evaluation."""

from . import util
from .crowd_detection_evaluator import CrowdDetectionEvaluator
from .detection_evaluator import DetectionEvaluator

__all__ = ["DetectionEvaluator", "CrowdDetectionEvaluator", "util"]



"""Set of functions to construct groups in a dataset, and compute analytics pre group
during e.g. evaluation
"""

from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from typing import Literal, overload

import numpy as np
import pandas as pd


@dataclass
class ContinuousGroup:
    """Data Class to encapsulate information to give to the cutting function of pandas
    as parameters, typically used to group continuous data by a limited number of
    groups, similarly to an histogram.

    Depending on the attributes, il will use either :func:`pandas.cut` or
    :func:`pandas.qcut` to give a particular label for each row of you dataframe
    """

    name: str
    """Name of the column to use the cutting function on"""

    bins: float | list[float] = 10
    """value given to the ``bin`` parameter of pandas' function. Can be either a float
    (for the number of bins), or a list of values that will be used as actual bins. Note
    that in the case of :func:`pandas.qcut`, only this attribute being a float makes
    sense."""

    qcut: bool = False
    """Whether to use :func:`pandas.qcut` or :func:`pandas.cut`. Qcut will design the
    bins so that each interval will contain the same number of samples, while cut will
    design the bins so that first and last bins are minimum and maximum value of
    considered column, and all the bins are equally spaced
    (similar to :func:`numpy.linspace`)
    """

    log: bool = False
    """When using cut (and not qcut), whether to separate bins equally in the linear
    space or the log space.
    As such, bins for lower values would be closer to each other"""

    label_type: str = "intervals"
    """What type of label to give to each group given by the cutting function.

    Can be either:
     - "intervals" (default): :class:`pandas.Interval` object usually given as Series
       values by :func:`pandas.cut` and :func:`pandas.qcut`
     - "mid": mid point between the two bins of each interval
     - "mean": mean value of data points comprised in a given interval
     - "median": median value of data points comprised in a given interval

     """

    def to_dict(self) -> dict[str, str | float | list[float] | bool]:
        """Serialize the ContinuousGroup object into a dictionary that can then be used
        as kwargs for :func:`.cut_group`

        Returns:
            Dictionary containing parameters to be read by :func:`.cut_group`
        """
        return {
            "group_name": self.name,
            "bins": self.bins,
            "log": self.log,
            "qcut": self.qcut,
            "label_type": self.label_type,
        }


group = str | ContinuousGroup
"""
Type alias to define a group

Group is either

- the name of a column (for discret groups, such as ``category_id``)
- a :class:`.ContinuousGroup` object to divide continuous data into a given number
  of groups, similar to histograms.

these parameters will be used for the function :func:`lours.util.grouper.cut_group`

Examples:
    Discret group::

        "size"


    Continuous group::

        continuousGroup(name="size", bins=10, log=False, qcut=True)

    Continuous group with bins::

        continuousGroup(name="size", bins=[0, 10, 20, 30], log=False, qcut=False)
"""

group_list = group | Sequence[group]
"""
Group list is either a group or an iterable of groups
"""


def cut_group(
    data: pd.Series | pd.DataFrame,
    group_name: str | None = None,
    bins: int | Iterable[float] = 10,
    label_type: str = "intervals",
    log: bool = False,
    qcut: bool = False,
) -> pd.Series:
    """Cut a dataframe according to one of its column values and criteria
    See :func:`pandas.cut`, :func:`pandas.qcut`

    Args:
        data: Dataframe to extract the column name from
        group_name: name of the column to extract
        bins: parameter used by both :func:`pandas.cut`, :func:`pandas.qcut`. Namely,
            it can be an int to describe the number of bins, or a list of floats, to
            either describe the actual bin edges for :func:`pandas.cut` or the quantile
            edges for :func:`pandas.qcut`
        label_type: what type of label to give to each group given by the cutting
            function.
            Can be either:

            - "intervals" (default): :class:`pandas.Interval` object usually given as
              Series values by :func:`pandas.cut` and :func:`pandas.qcut`
            - "mid": mid-point between the two bins of each interval
            - "mean": mean value of data points comprised in a given interval
            - "median": median value of data points comprised in a given interval

        log: Whether to use logarithmic scale or not, when bins is an integer.
            Useful when the values are not uniformly distributed. Defaults to False.
        qcut: Whether to use :func:`pandas.qcut` instead of :func:`pandas.cut`.
            See corresponding documentation for the differences.
            TL;DR, :func:`pandas.qcut` is based on quantiles (same number of occurrences
            in each bin) while :func:`pandas.cut` is based on values (same interval
            length for each bin). Defaults to False.

    Raises:
        ValueError: Raises an error when log option is selected but the extracted column
            has negative values

    Returns:
        Series with the same length as data, describing a mapping from id to bin. Bin
        labels are Interval Indices describing the upper and lower bound.
        See :class:`pandas.IntervalIndex`
    """
    if isinstance(data, pd.DataFrame):
        assert group_name is not None
        to_cut = data[group_name]
    else:
        to_cut = data
        if group_name is not None:
            to_cut.name = group_name
    cut_function = pd.qcut if qcut else pd.cut
    if (not log) or (not isinstance(bins, int)):
        if isinstance(bins, Iterable):
            bins = [*bins]
        result = cut_function(to_cut, bins)
    else:
        if to_cut.min() < 0:
            raise ValueError("Cannot use log on negative values")
        log_to_cut = np.log(to_cut)
        log_cut = cut_function(log_to_cut, bins)
        assert isinstance(log_cut, pd.Series)

        # Change labels to match actual values and not log ones

        def exp_labels(x: pd.Interval) -> pd.Interval:
            return pd.Interval(np.exp(x.left), np.exp(x.right))

        normal_labels = log_cut.cat.categories.map(exp_labels)
        result = log_cut.cat.rename_categories(normal_labels)
    if label_type == "intervals":
        return result
    elif label_type == "mid":
        return result.apply(lambda x: x.mid)
    elif label_type == "mean":
        means = to_cut.groupby(result, observed=False).mean()
        return result.apply(lambda x: means.loc[x])
    elif label_type == "median":
        means = to_cut.groupby(result, observed=False).median()
        return result.apply(lambda x: means.loc[x])
    raise ValueError("invalid label_type")


@overload
def make_pandas_compatible(
    data: pd.DataFrame, g: str
) -> tuple[str, str, Literal[True]]:
    pass


@overload
def make_pandas_compatible(
    data: pd.DataFrame,
    g: ContinuousGroup,
    root_data: pd.DataFrame | None = None,
    key_to_root: str = "image_id",
) -> tuple[str, pd.Series, Literal[False]]:
    pass


@overload
def make_pandas_compatible(
    data: pd.DataFrame,
    g: group,
    root_data: pd.DataFrame | None = None,
    key_to_root: str = "image_id",
) -> tuple[str, str | pd.Series, bool]:
    pass


def make_pandas_compatible(
    data: pd.DataFrame,
    g: group,
    root_data: pd.DataFrame | None = None,
    key_to_root: str = "image_id",
) -> tuple[str, str | pd.Series, bool]:
    """Construct group from :obj:`group` that will be used for pandas'
    `groupby <https://pandas.pydata.org/docs/reference/groupby.html>`__ method.

    - In the case it's only a name, keep it like that
    - Otherwise, we need to construct an index of data cut according to the given
      bins. This will create a :class:`pandas.Series` with categorical data

    Args:
        data: input DataFrame, must contain the column considered in group ``g``
        g: group depicting a column from ``data`` with potential bins.
            See :obj:`group`
        root_data: Potential root data where some ids in ``data`` refer to a particular.
            columns in ``root_data``. Defaults to None.
        key_to_root: column containing ``root_data`` row ids. Defaults to "image_id".

    Returns:
        Tuple with the 3 following values:

        1. group name
        2. group that can be understood by pandas'
           `groupby <https://pandas.pydata.org/docs/reference/groupby.html>`__ method.
           Can be a simple string referring to a column, or a :class:`pandas.Series`
           with categorical data
        3. boolean indicating whether the group is categorical
           (on which different values are independent of each other)
           or continuous (on which different values represent ranges of a continuous
           value, constructing a discretized histogram)
    """

    def construct_column_group_from_root(
        root_dataframe: pd.DataFrame, input_group_name: str
    ) -> pd.Series:
        column = root_dataframe.loc[data[key_to_root], input_group_name]
        column.index = data.index
        return column

    is_category = True
    if isinstance(g, str):
        group_name = g
        if group_name in data.columns:
            group = g
        else:
            assert root_data is not None
            assert group_name in root_data.columns
            group = construct_column_group_from_root(root_data, g).astype("category")
    else:
        group_name = g.name
        if group_name in data.columns:
            group = cut_group(data, **g.to_dict())  # pyright: ignore
        else:
            assert root_data is not None
            assert group_name in root_data.columns
            group = cut_group(
                construct_column_group_from_root(root_data, group_name),
                **g.to_dict(),  # pyright: ignore
            )
        is_category = False
    return group_name, group, is_category


def get_group_names(groups: group_list) -> list[str]:
    """From a list of groups, get the list of associated names.

    Args:
        groups: single group lor Sequence of groups to extract the names from.

    Returns:
        Names of given groups.
    """
    return [g if isinstance(g, str) else g.name for g in groups_to_list(groups)]


def groups_to_list(groups: group_list) -> list[group]:
    """Convert a single group or Sequence of groups to a list of groups
    (possibly with one element)

    Args:
        groups: Sequence of groups or single groups to convert

    Returns:
        Actual list of groups, more easily handled by other functions.
    """
    if isinstance(groups, str | ContinuousGroup):
        return [groups]
    else:
        return list(groups)


def group_relational_data(
    input_data: pd.DataFrame,
    groups: group_list,
    root_data: pd.DataFrame | None = None,
    key_to_root: str = "image_id",
) -> tuple[dict[str, str | pd.Series], list[str], list[str]]:
    """Create groups that will be applied on ``input_data`` with the
    :meth:`pandas.DataFrame.groupby` method. can be used with a ``root_data`` relational
    DataFrame containing values that we might want to group, provided ``input_data``
    contains a column with reference to a row in ``root_data``.

    Args:
        input_data: DataFrame to group
        groups: groups to apply to ``input_data`` or `root_data``.
            Can be a simple string in the case of categorical data, or a dictionary.
            See :obj:`.group`.
        root_data: DataFrame containing information ``input_data`` may refer to.
            Defaults to None.
        key_to_root: column name in ``input_data`` for the key to ``root_data``.
            Defaults to "image_id".

    Returns:
        3 different objects are returned:

        1. A dictionary with the created groups and their name as a key. The groups can
            be directly used in a `input_data.groupby` call
        2. A list of all category groups, where different values are independent
            from each other
        3. A list of all continuous groups, on which different values represent ranges
            of a continuous value, constructing a discretized histogram

        Note that the two list together should be as long as the group dictionary, and
        their elements must refer to all the actual keys of the dictionary.
    """
    groups = groups_to_list(groups)

    groups_dict: dict[str, str | pd.Series] = {}
    category_groups: list[str] = []
    continuous_groups: list[str] = []
    for g in groups:
        name, group, is_category = make_pandas_compatible(
            input_data, g, root_data, key_to_root
        )
        if is_category:
            category_groups.append(name)
        else:
            assert isinstance(group, pd.Series)
            continuous_groups.append(name)
        groups_dict[name] = group

    return groups_dict, category_groups, continuous_groups



import warnings
from pathlib import Path

import pandas as pd
from typing_extensions import Self

from ..dataset import Dataset
from ..utils.label_map_merger import merge_label_maps
from ..utils.parquet_saver import dict_from_parquet, dict_to_parquet
from ..utils.testing import assert_frame_intersections_equal


class Evaluator:
    """Abstract class of Evaluator, made to measure prediction quality with respect to a
    Dataset of groundtruth annotations. Depending on data type, the method used for
    evaluation might differ, refer to the specialized classes for information. The
    fundamental building block is the Dataset object representing the groundtruth.
    additional kwargs given to the constructor are also Dataset objects that must match
    the groundtruth, in terms of image and label maps (if any)
    """

    name: str | None
    """Name of Evaluator. Can be deduced from groundtruth's dataset name and will be
    used in export functions like :meth:`.DetectionEvaluator.to_fiftyone`"""

    groundtruth: pd.DataFrame
    """DataFrame comprising annotation data. Must have at least ``image_id`` column"""

    predictions_dictionary: dict[str, pd.DataFrame]
    """dictionary of DataFrames comprising prediction data. Must have at least
    ``image_id`` and ``confidence`` columns"""

    images: pd.DataFrame
    """DataFrame comprising image data. This dataframe should be referred to by both gt
    and predictions with the ``image_id`` column"""

    images_root: Path
    """Root folder where to grab images. Image filepath will be concatenation of
    images_root and their relative path"""

    label_map: dict[int, str]
    """Mapping from category_id to category_str. If used, is generally taken from the
    groundtruth Dataset. The prediction must be compatible with it"""

    def __init__(
        self,
        groundtruth: Dataset,
        name: str | None = None,
        **predictions: Dataset,
    ):
        """Constructor of the Evaluator object.

        Args:
            groundtruth: Dataset object representing the ground truth with
                annotations, image data and label_map
            name: Name of Evaluator. If set to None, will be deduced from groundtruth's
                dataset name
            **predictions: keyword arguments for additional datasets to compare the
                groundtruth to. Its images must match the groundtruth dataset (see
                add_prediction_dataset method below).
        """
        if name is None:
            self.name = groundtruth.dataset_name
        else:
            self.name = name
        self.images_root = groundtruth.images_root
        self.groundtruth = groundtruth.annotations
        self.images = groundtruth.images.drop("split", axis=1, errors="ignore")
        self.label_map = groundtruth.label_map
        self._default_annotation_columns_with_types = (
            groundtruth._default_annotation_columns_with_types
        )
        self._default_image_columns_width_types = (
            groundtruth._default_image_columns_with_types
        )
        self.predictions_dictionary = {}
        for predictions_name, predictions_df in predictions.items():
            self.add_predictions_dataset(predictions_name, predictions_df)

    def get_image_attributes(self) -> list[str]:
        """Get the name of columns related to image attributes. In other words, get
        columns that are NOT the default ones.

        The actual attribute values can then be
        ``self.images[self.get_image_attributes()]``

        Returns:
            list of column names in ``self.images`` that represent attributes
        """
        return [
            str(c)
            for c in self.images.columns
            if c not in self._default_image_columns_width_types.keys()
        ]

    def get_annotations_attributes(
        self, predictions_name: str | None = None
    ) -> list[str]:
        """Get the name of columns related to annotations attributes. In other words,
        get columns that are NOT the default ones.

        the actual attribute values can then be

        .. code-block:: python

            self.predictions_dictionary[predictions_name][
                self.get_annotations_attributes()
            ]

        Args:
            predictions_name: name of predictions to extract not default column from.
                If None, will use ``self.groundtruth``. Defaults to None.

        Returns:
            list of column names in ``self.annotations`` that represent attributes
        """
        if predictions_name is None:
            predictions = self.groundtruth
        else:
            predictions = self.predictions_dictionary[predictions_name]
        return [
            str(c)
            for c in predictions.columns
            if c not in self._default_annotation_columns_with_types.keys()
        ]

    def add_predictions_dataset(self, predictions_name: str, predictions: Dataset):
        """Method to add predictions to the Evaluator from a Dataset object.
        The prediction dataset must match the Evaluator data:

        - prediction label_map must be equal or a subset to the
            Evaluator's label map
        - image data must be the same, except the relative path
            (it can change although the image has not) i.e. there must be
            the same number and ids of images and all the columns in the prediction
            image data must match the corresponding ones
            in the evaluator image data.

        Note that this method will overwrite a potentially already existing prediction
        dataframe

        Args:
            predictions_name: name of predictions to add. It will then be used as key in
                the ``self.predictions_dictionary`` attribute.
            predictions: prediction Dataset, from which the annotations will
                be extracted and added to the evaluator.
        """
        assert "confidence" in predictions.annotations, "Not a prediction dataset"

        new_label_map = merge_label_maps(
            self.label_map, predictions.label_map, method="outer"
        )
        if new_label_map != self.label_map:
            warnings.warn(
                f"Although compatible, '{predictions_name}' prediction label map is"
                " larger than groundtruth label map",
                RuntimeWarning,
            )
            self.label_map = new_label_map

        if not predictions.images.index.isin(self.images.index).all():
            raise ValueError(
                "Some image ids in given predictions are not present in the evaluator"
                " image index"
            )

        try:
            assert_frame_intersections_equal(self.images, predictions.images)
        except AssertionError as e:
            raise AssertionError(
                "Groundtruth and Prediction images are not consistent on their"
                " overlapping indices and columns. You might want to consider the"
                " Dataset.reindex() method."
            ) from e

        self.add_predictions(predictions_name, predictions.annotations)

    def add_predictions(self, predictions_name: str, predictions: pd.DataFrame):
        """Method to add predictions to the Evaluator from a dataframe.
        No check will be done on image data the annotations refer to. However, it will
        check that ``image_id`` values of ``predictions`` are contained in the
        evaluator's ``image_data`` and ``category_id`` values are contained in the
        label map

        Note that this method will overwrite a potentially already existing prediction
        dataframe

        Args:
            predictions_name: name of predictions to add. It will then be used as key in
                the ``self.predictions_dictionary`` attribute.
            predictions: prediction dataframe to be added to
                the evaluator.
        """
        predictions_image_ids = set(predictions["image_id"].unique())
        assert set(self.images.index).issuperset(predictions_image_ids)
        predictions_class_ids = set(predictions["category_id"].unique())
        assert predictions_class_ids.issubset(self.label_map.keys())
        self.predictions_dictionary[predictions_name] = predictions

    def to_parquet(self, output_dir: Path | str, overwrite: bool = False) -> None:
        """Save the current object to a folder containing parquet files for dataframes
        inside this object, and a metadata.yaml file for other attributes.

        Args:
            output_dir: output directory where the files will be created.
                If ``overwrite`` is set to False, it must not already exist.
            overwrite: if set to True, will remove the directory at ``output_dir``
                if it already exists. Defaults to False.
        """
        dict_to_parquet(
            {k: v for k, v in vars(self).items() if not k.startswith("_")}
            | {"__name__": self.__class__.__name__},
            Path(output_dir),
            overwrite=overwrite,
        )

    @classmethod
    def from_parquet(cls, input_dir: Path | str) -> Self:
        """Class method to construct an instance of this class or a subclass.
        the parquet folder must have been created with the method ``to_parquet``
        (see above)

        Args:
            input_dir: Path to directory containing the metadata.yaml file along with
                the different parquet files

        Raises:
            ValueError: Raised when the object name contained in
                ``input_dict['__name__']`` is not the same as the name of the class this
                method is called from. For example, you can't call
                :meth:`.Evaluator.from_parquet` with a folder created by a
                :class:`DetectionEvaluator` object.

        Returns:
            New object of the same subclass as the method is caled from,
            containing data loaded from the parquet files in the input directory
        """
        input_dict = dict_from_parquet(Path(input_dir))
        if cls.__name__ != input_dict["__name__"]:
            raise ValueError(
                f"Wrong object type for parquet archive. Expected {cls.__name__}, got"
                f" {input_dict['__name__']}"
            )
        groundtruth_dataset = Dataset(
            images_root=input_dict["images_root"],
            images=input_dict["images"].assign(split=None),
            annotations=input_dict["groundtruth"],
            label_map=input_dict["label_map"],
        )
        predictions = input_dict["predictions_dictionary"]
        evaluator = cls(groundtruth_dataset)
        for name, predictions in predictions.items():
            evaluator.add_predictions(name, predictions)
        for k, v in vars(evaluator).items():
            loaded_value = input_dict.get(k, None)
            if loaded_value is not None:
                evaluator.__dict__[k] = loaded_value
        return evaluator

    def _ipython_display_(self):
        """Function to display the Dataset as an HTML widget when using notebooks"""
        import ipywidgets as widgets
        from IPython.display import display

        tab = widgets.Tab()

        descr_str = (
            "<b> Evaluation object, containing "
            f"{len(self.images):,} images, {len(self.groundtruth):,} groundtruth "
            f"objects, and {len(self.predictions_dictionary)} prediction sets </b>"
        )

        title = widgets.HTML(descr_str)

        components_widgets = self._get_widgets()

        tab.children = [*components_widgets.values()]
        tab.titles = [*components_widgets.keys()]

        display(widgets.VBox([title, tab]))

    def _get_widgets(self):
        import ipywidgets as widgets
        from IPython.display import display

        label_map_df = pd.Series(self.label_map, name="category string").to_frame()
        label_map_df.index.name = "categorty_id"

        # create output widgets
        widget_images = widgets.Output()
        widget_groundtruth = widgets.Output()
        widget_predictions = {
            p_name: widgets.Output() for p_name in self.predictions_dictionary
        }
        widget_label_map = widgets.Output()

        # render in output widgets
        with widget_images:
            display(self.images)
        with widget_groundtruth:
            display(self.groundtruth)
        for p_name, p in self.predictions_dictionary.items():
            with widget_predictions[p_name]:
                display(p)
        with widget_label_map:
            display(label_map_df)

        return {
            "Images": widget_images,
            "Groundtruth": widget_groundtruth,
            **widget_predictions,
            "label_map": widget_label_map,
        }



from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Any

import imagesize
import pandas as pd

from lours.dataset import Dataset

from ...utils.bbox_converter import column_names_from_format_string, import_bbox

IMG_FORMATS = "bmp", "dng", "jpeg", "jpg", "mpo", "png", "tif", "tiff", "webp", "pfm"


def construct_label_map(annotations: pd.DataFrame) -> dict[int, str]:
    """Construct label map from annotation DataFrame, with ``category_id`` and
    ``category_str`` columns. Get all category string associated with each category id.
    Normally, there should be only one per id

    Args:
        annotations: DataFrame containing category id and category name information.
            Should contain at least ``category_id`` and  ``category_str`` columns.

    Raises:
        ValueError: Inconsistency in category ids and names. The ``id -> name`` mapping
            should be bijective.

    Returns:
        dictionary containing label map, with category id as key, and category name
        as value
    """
    label_map_df = (
        annotations[["category_id", "category_str"]]
        .groupby("category_id")["category_str"]
        .unique()
    )
    label_map_check = label_map_df.apply(len) != 1
    if label_map_check.any():
        print("Problem with label map, some category ids have multiple different names")
        print(label_map_df[label_map_check])
        raise ValueError("Invalid label map")
    label_map = {k: v[0] for k, v in label_map_df.items()}
    # Raise an error if two ids have the same category name
    if len(set(label_map.values())) != len(label_map):
        print("Problem with label map, some category names are present in multiple ids")
        print(label_map)
        raise ValueError("Invalid label map")
    return label_map  # pyright: ignore


def convert_str(string: str) -> str | int | float:
    """String converter tool to read a file, parse and automatically convert the string
    to integer or float if possible. Will first try to convert to int, then float,
    then will return as is.

    Args:
        string: string containing information to be parsed

    Returns:
       converted string, in the most convenient format
    """
    try:
        result = float(string)
        if result == int(result):
            return int(result)
        else:
            return result
    except ValueError:
        return string


def get_relative_image_path(dataset_path: Path, image_path: Path | str) -> Path:
    """Tool function to get relative path between dataset_path and image_path, which
    might be absolute. Used to populate the ``relative_path`` in the images dataframe
    of the dataset or evaluator object, which should check the fact that
    ``dataset.images_root / relative_path`` should always lead to a valid image file

    Args:
        dataset_path: root path of considered dataset
        image_path: image path of a particular image. May be absolute, and need
            to be converted to be relative the dataset_path

    Raises:
        ValueError: image_path is not included in dataset_path. Probably means the
            dataset path is too specific and should be higher in file hierarchy

    Returns:
        Converted image path to be relative to dataset path
    """
    image_path = Path(image_path)
    if image_path.is_absolute():
        try:
            return image_path.relative_to(dataset_path)
        except ValueError as e:
            raise ValueError(
                "Image paths are not contained in given dataset folder. "
                "If you want to use absolute path, "
                "try giving '/' to dataset_path argument"
            ) from e
    else:
        return image_path


def get_images_from_folder(
    folder_path: Path, img_formats: Iterable[str] = IMG_FORMATS
) -> list[Path]:
    """Function to scrape all images in a folder, starting from a list of img formats

    Args:
        folder_path: where to search images
        img_formats: list of file extensions to consider during the globbing

    Returns:
        list of all paths leading to an image with the desired extensions.
    """
    output = []
    for img_format in img_formats:
        output.extend(folder_path.glob(f"**/*.{img_format}"))
    return [path.relative_to(folder_path) for path in output]


def get_image_info(
    image_number: int,
    relative_path: Path,
    absolute_path: Path | None,
    image_info: pd.DataFrame | None = None,
) -> dict[str, Any]:
    """Get image information, either from image info dataframe or from image itself,
    getting the image dimension by reading its header

    Args:
        image_number: number of image in the file list. If image_info is not
            available, will be used for image id
        relative_path: path that will be used to find the image in the image_info
            dataframe, if given
        absolute_path: absolute to load the image data directly from the file. Can be
            None if image_info has an entry with the same ``relative_path`` value
        image_info: DataFrame including image size and image id to match the ids of
            another dataset for example. Must have at least ``relative_path``,
            ``width`` and ``height`` columns. Defaults to None.

    Returns:
        dictionary with width height and id keys
    """
    current_image = {"relative_path": relative_path, "type": relative_path.suffix}
    if image_info is None or relative_path not in image_info["relative_path"].values:
        if absolute_path is None:
            raise ValueError(
                "You must at least provide valid image_info or absolute_path"
            )
        width, height = imagesize.get(absolute_path)
        current_image["width"] = width
        current_image["height"] = height
        image_id = image_number
        current_image["id"] = image_id
    else:
        current_image_info = image_info[
            image_info["relative_path"] == relative_path
        ].iloc[0]
        current_image["width"] = current_image_info["width"]
        current_image["height"] = current_image_info["height"]
        image_id = current_image_info.name
        current_image["id"] = image_id
    return current_image


def to_dataset_object(
    images_root: Path,
    label_map: dict[int, str] | None,
    images: Sequence[dict],
    annotations: Sequence[dict],
    box_format: str = "cxwcyh",
    ids_map: dict[int, dict[str, Any]] | None = None,
) -> Dataset:
    """Create the dataset object from aggregated lists of dictionaries

    Args:
        images_root: path where the images are located and from where relative paths
            are given
        label_map: dictionary of category id vs category name
        images: list of image dictionaries. Each dictionary is one image
        annotations: list of annotations dictionaries
        box_format: expected type of box format. See :mod:`lours.utils.bbox_converter`
            Defaults to "cxwcyh"
        ids_map: dictionary to remap classes back to their original id values.
            This is a special case of darknet where the ids are almost always changed
            because they need to be sequential

    Returns:
        created dataset objects with the right category ids
    """
    if images:
        images_df = pd.DataFrame(images).set_index("id")
    else:
        assert (
            len(annotations) == 0
        ), "Got empty image sequence, but a non empty annotation sequence"
        images_df = None
    if annotations:
        annotations_df = pd.DataFrame(annotations)
        if "id" in annotations_df.columns:
            annotations_df = annotations_df.set_index("id")
        else:
            annotations_df.index.name = "id"
    else:
        annotations_df = None
    if images_df is not None and annotations_df is not None:
        bboxes = import_bbox(annotations_df, images_df, input_format=box_format)
        columns_to_drop = column_names_from_format_string(box_format)
        annotations_df = annotations_df.drop(
            columns_to_drop,
            axis=1,
        )
        annotations_df = pd.concat([annotations_df, bboxes], axis=1)
        if label_map is None:
            label_map = construct_label_map(annotations_df)
    dataset = Dataset(
        images_root,
        images_df,
        annotations_df,
        label_map,
    )
    if ids_map is not None:
        id_remapping = {int(i): line["id"] for i, line in ids_map.items()}
        new_label_map = {line["id"]: line["name"] for line in ids_map.values()}
        dataset = dataset.remap_classes(id_remapping, new_label_map)
    return dataset


def parse_annotation_name(
    annotations_file_path: str | Path,
    split_name_mapping: dict[str, list[str]] | None = None,
) -> tuple[str | None, str | None]:
    """Deduce name of dataset and name of split by assuming it is in the form
    '<dataset_name>_<split_name>.<extension>'

    For example, 'coco_train.json' will be parsed to return 'coco' and 'train'

    Args:
        annotations_file_path: name of the annotation file without extension or path to
            the annotation file which name will be parsed.
        split_name_mapping: Dictionary with split names you want to appear in the lours
            dataset as keys and a list of possible words you want this name to replace
            as values. For example, remap split names abbreviations to their full name
            so that "val" becomes "validation". If set to None, will simply map
            variations of 'train', 'valid', 'eval' to them, i.e. 'training' gets
            replaced by 'train', 'val' and 'validation' get replaced by 'valid' and
            'evaluation' and 'test' get replaced by 'eval'. Defaults to None.

    Returns:
        tuple containing two names : dataset name and split name. They can be none in
        the case parsing was not successful.

    Example:
        >>> parse_annotation_name("my_dataset_test")
        ('my_dataset', 'eval')
        >>> parse_annotation_name("my_dataset_hello", {"hey": ["hello", "hi"]})
        ('my_dataset', 'hey')
        >>> parse_annotation_name("my_dataset")
        ('my', 'dataset')
        >>> parse_annotation_name("mydataset")
        ('mydataset', None)
    """
    if split_name_mapping is None:
        split_name_mapping = {
            "train": ["train", "training"],
            "valid": ["val", "valid", "validation"],
            "eval": ["eval", "evaluation", "test"],
        }
    if isinstance(annotations_file_path, Path):
        name = annotations_file_path.stem
    else:
        name = annotations_file_path
    if "_" in name:
        parsed_name, parsed_split = name.rsplit("_", maxsplit=1)
        for possible_split, possible_names in split_name_mapping.items():
            if parsed_split in possible_names:
                parsed_split = possible_split
        return parsed_name, parsed_split
    return name, None
